
#include <AL/al.h>
#include <AL/alut.h>
/////////////////////////////////////////////////////////////////
class OpenALSoundSource : public GSE_Positional, 
			  public GSE_Directional
{
  friend class OpenALSoundSystem;
 private:
  ALuint      m_ALSource;
  int	      m_bLooping;
  float	      m_fGain;
 private:
  ////////////////////
  ~OpenALSoundSource()
  {
    Destroy();
  }
  ////////////////////
  OpenALSoundSource() : m_fGain(1.0f)
  {
    Init();
  }
  ////////////////////
  void Init()
  {
    ALenum error;
    alGenSources( 1, &m_ALSource );
    if ( (error = alGetError()) != AL_NO_ERROR )
    {
      cerr << "alGenSources error" << error << endl;
      exit(1);
    }
  }
  ////////////////////
  void Destroy()
  {
    alDeleteSources( 1, &m_ALSource );
  }
  ////////////////////
  void AttachBuffer( ALuint alBuffer )
  {
    ALenum error;
    alSourcei( m_ALSource, AL_BUFFER, alBuffer );
    if ( (error = alGetError()) != AL_NO_ERROR )
    {
      cerr << "alGenSources error" << error << endl;
      exit(1);
    }
  }

 public:
  ////////////////////
  void Play()
  {
    alSourceRewind(m_ALSource);
    alSourcePlay(m_ALSource);
  }
  ////////////////////
  void Pause()
  {
    alSourcePause(m_ALSource);
  }
  ////////////////////
  void Stop()
  {
    alSourceStop(m_ALSource);
  }
  void SetPosition( const GSE_Vector3 &vPosition )
  {
    GSE_Positional::SetPosition( vPosition );
    alSourcefv(  m_ALSource, AL_POSITION, vPosition.m_pValues );
  }
  void SetDirection( const GSE_Vector3 &vDirection )
  {
    GSE_Directional::SetDirection( vDirection );
    alSourcefv( m_ALSource, AL_DIRECTION, vDirection.m_pValues );
  }
  void SetLooping( int bFlag )
  {
    m_bLooping = bFlag;
    alSourcei( m_ALSource, AL_LOOPING, m_bLooping ? AL_TRUE : AL_FALSE );
  }
  int IsLooping()
  {
    return m_bLooping;
  }
  void SetGain( float fGain ) 
  {
    m_fGain = fGain;
    alSourcef( m_ALSource, AL_GAIN, m_fGain );
  }
};

/////////////////////////////////////////////////////////////////
#define MAX_AL_BUFFERS 12
class OpenALSoundSystem 
{
 private:  
  ALuint m_arrBuffers[MAX_AL_BUFFERS];
  vector<OpenALSoundSource *> m_vecSources;
 public:
  ////////////////////
  OpenALSoundSystem()
  {
    Init();
  }
  ////////////////////
  int LoadSound( unsigned int nSlot, const string &strFilename )
  {
    ALenum error;
    m_arrBuffers[nSlot] = alutCreateBufferFromFile( strFilename.c_str());
    if ( ( error = alGetError()) != AL_NO_ERROR )
    {
      cerr << "alutCreateBufferFromFile failed " << error << endl;
      exit(1);
    }
    return 0;
  }
  ////////////////////
  OpenALSoundSource * CreateSource( unsigned int nBuffer )
  {
    OpenALSoundSource *pSource = new OpenALSoundSource();
    pSource->AttachBuffer( m_arrBuffers[nBuffer] );
    return pSource;
  }
  ////////////////////
  void SetListener( const GSE_Orientable & orientable, const GSE_Positional & positional )
  {
    float afOrientable[6];
    memcpy( &(afOrientable[0]), orientable.GetForwardVector().m_pValues, sizeof(float)*3 );
    memcpy( &(afOrientable[3]), orientable.GetUpVector().m_pValues,      sizeof(float)*3 );
    alListenerfv(AL_POSITION, positional.GetPosition().m_pValues);
    alListenerfv(AL_ORIENTATION, afOrientable);
  }
 private:

  ////////////////////
  void Init()
  {
    ALenum error;
    alutInit(0, NULL);
    alGetError();
    alGenBuffers( MAX_AL_BUFFERS, m_arrBuffers);

    if ( (error = alGetError()) != AL_NO_ERROR )
    {
      cerr << "OpenAL Error: " << error << endl;
      exit(1);
    }
    ALint iNumChannels;
    alGetBufferi( m_arrBuffers[0], AL_CHANNELS, &iNumChannels );
    cerr << "1st buffer has " << iNumChannels << " channels" << endl;
  }
  ////////////////////
  void Destroy()
  {
    alDeleteBuffers(MAX_AL_BUFFERS, m_arrBuffers);
    m_vecSources.clear();
    alutExit();
  }
};
/////////////////////////////////////////////////////////////////
