/////////////////////////////////////////////////////////////////
///  Very crude demo of GSE features.
///  Copyleft Anssi Grhn/eNtity 2007.
///  This program is GPL v2 licenced.
/////////////////////////////////////////////////////////////////
#include "GSE.h"   // include GSE headers
#include <string>  
#include <map>
#include <iostream>   
#include "SoundSystem.h" // Crude OpenAL sound system
/////////////////////////////////////////////////////////////////
// These make code more readable...
using std::string;
using std::map;
using std::list;
using std::cerr;
using std::endl;
///////////////////////////////////////////////////////////////////
/// Animation sequence. Defines a range of frames, which can be looped.
/////////////////////////////////////////////////////////////////
class AnimationSequence
{
public:
  // Start frame
  unsigned int m_nStart;
  // End frame
  unsigned int m_nEnd;
  // Is this looping 
  unsigned int m_bLooping;
  // Constructor
  AnimationSequence( unsigned int nStart, 
		     unsigned int nEnd,
		     int bLooping)
  {
    m_nStart = nStart;
    m_nEnd   = nEnd;
    m_bLooping = bLooping;
  }
};
/////////////////////////////////////////////////////////////////
// Particle VelocityInitializer. This is used in the particle system
// which creates the steaming breath effect.
// It provides some variation the initial speed of particle at birth. 
template <class ParticleType>
class VelocityInitializer 
{
public:
  GSE_Vector3 m_vVelocity;
  VelocityInitializer() {}
  inline void operator()( ParticleType &p ) const 
  {
    p.m_vVelocity = m_vVelocity * Math::GetRandom<float>(5.004f, 8.03f);
    
  }
};
/////////////////////////////////////////////////////////////////
// This policy makes the particles rise up towards the sky.
// Also used in steaming breath effect.
template <class ParticleType>
class SkyforceAction
{
protected:
  float m_fPassedTimeInSec;
public:
  SkyforceAction() : m_fPassedTimeInSec(0.0) {}
  inline void Prepare( unsigned int nPassedTimeInMS )
  {
    m_fPassedTimeInSec = nPassedTimeInMS * 0.001f;
  }
  inline void operator()( ParticleType &p) const 
  {
    if ( p.m_vVelocity[GSE_Vector3::X] > 0.0f )
    {
      float fReduction = 0.4f;
      p.m_vVelocity[GSE_Vector3::X] -= fReduction * m_fPassedTimeInSec;
    }
    else if ( p.m_vVelocity[GSE_Vector3::X] < 0.0f )
    {
      float fReduction = 0.4f;
      p.m_vVelocity[GSE_Vector3::X] += fReduction * m_fPassedTimeInSec;
    }
    if ( p.m_vVelocity[GSE_Vector3::Z] > 0.0f )
    {
      float fReduction =  0.4f;
      p.m_vVelocity[GSE_Vector3::Z] -= fReduction * m_fPassedTimeInSec;
    }
    else if ( p.m_vVelocity[GSE_Vector3::Z] < 0.0f )
    {
      float fReduction = 0.4f;
      p.m_vVelocity[GSE_Vector3::Z] += fReduction * m_fPassedTimeInSec;
    }
    p.m_vVelocity[GSE_Vector3::Y] = 4.0f*m_fPassedTimeInSec;
  }
};
/////////////////////////////////////////////////////////////////
/// This policy class fades the steam clouds as time passes,
/// by changin the alpha values of particles.
template <class ParticleType>
class ColorAction
{
protected:
  float m_fPassedTimeInSec;
public:
  ColorAction() : m_fPassedTimeInSec(0.0) {}
  inline void Prepare( unsigned int nPassedTimeInMS )
  {
    m_fPassedTimeInSec = nPassedTimeInMS * 0.001f;
  }
  inline void operator()( ParticleType &p) const 
  {
    p.m_fEnergy = p.m_vColor[GSE_Color::A] -= 0.1f * m_fPassedTimeInSec;
  }
};
/////////////////////////////////////////////////////////////////
/// This policy enlarges the steam clouds as time passes.
/// They tend to spread, don't they?
template <class ParticleType>
class GSE_SizeAction
{
 private:
  float m_fPassedTimeInSec;
 public:
  float m_fSize;

  GSE_SizeAction() : m_fSize(1.0f){}
  inline void Prepare( unsigned int nPassedTimeInMS )
  {
    m_fPassedTimeInSec = (float)nPassedTimeInMS * 0.001f;
  }
  inline const void operator()(ParticleType &p) const
  {
    p.m_fSize += m_fSize * m_fPassedTimeInSec;
  }
}; // end of GSE_SizeInitializer
/////////////////////////////////////////////////////////////////
/// Here we define type for our particle system. It has at most 
/// 100 particles at a time.
/// It looks pretty dreadful with all those policies defined.
/// But later on, we refer to this as CSteamParticleSystem.
typedef GSE_ParticleSystem<100, 
			   GSE_CompletePolicy<GSE_Particle, 
					      GSE_ColorInitializer<GSE_Particle>,
					      GSE_SizeInitializer<GSE_Particle>, 
					      VelocityInitializer<GSE_Particle>, 
					      GSE_EnergyInitializer<GSE_Particle>, 
					      GSE_NullPolicy<GSE_Particle> >,
			   GSE_CompletePolicy<GSE_Particle, 
					      ColorAction<GSE_Particle>,
					      GSE_SizeAction<GSE_Particle>, 
					      SkyforceAction<GSE_Particle>, 
					      GSE_NullPolicy<GSE_Particle>, 
					      GSE_MoveAction<GSE_Particle> >, 
			   GSE_Particle> CSteamParticleSystem;
/////////////////////////////////////////////////////////////////
/// This is a scenegraph node for the particle system.
/// Using this node, we can attach it to proper place in respect to
/// the drawf.
class CSteamingBreath : public GSE_SceneParticleSystemNode
{
  /// We need to allow scenegraph to call our constructor,
  friend class GSE_SceneGraph;
protected:
  /// Here is the actual particle system.
   CSteamParticleSystem m_Breath;  
  // This enforces the nodes to be included in a graph.
   CSteamingBreath() { }
  ~CSteamingBreath() {}
public:
  ////////////////////
  /// This method will load textures and all other stuff which is required.
  void Construct()
  {
    /// Load image ./data/breath.jpg, use alpha channel, 
    /// make 2D texture out of it and optimize for performance, if possible.
    GSE_Texture *pTexture = GSE_OglTextureMgr::GetInstance()->CreateTextureFromImage("./data/breath.jpg", 1, 
										     TextureType::GSE_TEX2D, 
										     TextureQuality::PERFORMANCE);

    GSE_AnimParticleSystemNode *pAnimNode = NULL;
    GSE_TextureNode *pTextureNode = NULL;
    ////////////////////
    /// We use additive blending for these particles.
    GSE_BlendingStateNode *pBlendState = GetGraph()->GetRenderGraph().CreateNode<GSE_BlendingStateNode>();
    
    GSE_ParticleSystemNode *pParticleSystemNode = NULL;
    pBlendState->SetEnabled(1);
    pBlendState->SetSrcOp( GL_SRC_ALPHA );
    pBlendState->SetDestOp( GL_ONE);
    ////////////////////
    ///  This is the force which draws particle towards sky.
    GSE_Vector3 vSkyForce(0.0, 3.0f, 0.0f);
    m_Breath.m_InitializePolicy.m_ColorPolicy.m_vColor.Set(1,1,1,0.5f);
    m_Breath.m_InitializePolicy.m_SizePolicy.m_fSize = 5.0f;
    m_Breath.m_ActionPolicy.m_SizePolicy.m_fSize = 2.0f;
    /// Here we create necessary structures for adding it to scene graph
    /// and animatin graph.
    GSE_ParticleSystemConverter::Convert( &m_Breath, pTexture, 
					  &pTextureNode, 
					  &pParticleSystemNode,
					  &pAnimNode,
					  &GetGraph()->GetAnimationGraph(),
					  &(GetGraph()->GetRenderGraph()));
    assert(pAnimNode!=NULL);
    /// This texture is partially transparent, so we need to hint renderer.
    pTexture->SetTransparentHint( 1 );

    GSE_ColorNode *pWhiteColor = GetGraph()->GetRenderGraph().CreateNode<GSE_ColorNode>();
    pWhiteColor->Set(1,1,1,1);

    /// We need to disable depth buffer writes, when rendering these particles.
    GSE_DepthBufferStateNode *pDepthTestOnWritesDisabled = GetGraph()->GetRenderGraph().CreateNode<GSE_DepthBufferStateNode>();
    pDepthTestOnWritesDisabled->SetDepthBufferWriting(0);
    pDepthTestOnWritesDisabled->SetDepthTest(1);

    /// We animate this particle system by adding it to animation graph.
    GetGraph()->GetAnimationRoot()->AddEdge( pAnimNode );

    // The first node of render chain is set to rendernode of this scenegraph 
    // node.
    SetRenderNode( pBlendState );
    /// Rest of that chain is built here.
    pBlendState->AddEdge( pDepthTestOnWritesDisabled );
    pDepthTestOnWritesDisabled->AddEdge(pWhiteColor);
    pWhiteColor->AddEdge( pTextureNode );
    SetRenderParticleSystemNode( pParticleSystemNode);
    assert(pParticleSystemNode->GetGeometryData() != NULL);
    
    
  }
  ////////////////////
  /// This method creates a new breath, so to speak.
  /// It makes the particle system blow out one particle more.
  void Breathe( const GSE_Vector3 & vPosition, const GSE_Vector3 &vVelocity )
  {
    m_Breath.m_InitializePolicy.m_VelocityPolicy.m_vVelocity = vVelocity;
    m_Breath.Init( 1, vPosition );
  }
};
///////////////////////////////////////////////////////////////////
/// This class is our main character class.
/// It represents a dwarf. A dwarf can run, hit, and breathe.
class CDwarf : public GSE_SceneObjectNode
{
  friend class GSE_SceneGraph;

protected:
  /// Dwarf is animated, and we need animation sequences to represent it.
  map< string, AnimationSequence> m_AnimationSequences;
  /// Animation speed.
  float m_fAnimationFPS;
  /// This is the particlesystem which generates breath effect.
  CSteamingBreath *m_pBreath;
  /// This is the main animation node for Dwarf.
  GSE_JointNode *m_pDwarfRootJoint;
public:
  /// This node is used to grab weapon into right hand, ie. weapon is attached to this node.
  GSE_SceneLimbNode *m_pWeaponGrabRight;
  /// This node is used to grab weapon into left hand, ie. weapon is attached to this node.
  GSE_SceneLimbNode *m_pWeaponGrabLeft;
  /// This node is the head node.
  GSE_SceneLimbNode *m_pHead;
protected:
  /// This points to the weapon in right hand.
  GSE_SceneObjectNode *m_pWeaponRight;
  /// This points to the weapon in left hand.
  GSE_SceneObjectNode *m_pWeaponLeft;
  /// This matrix defines the transform, which makes weapon appear in correct orientation in right hand.
  GSE_Matrix4x4f m_mRightHandTransform;  
  /// This matrix defines the transform, which makes weapon appear in correct orientation in left hand.
  GSE_Matrix4x4f m_mLeftHandTransform;  
  /// This is true, if right hand is occupied.
  unsigned int m_bRightHandFull;
  /// This is true, if left hand is occupied.
  unsigned int m_bLeftHandFull;
  ////////////////////
  /// Constructor.
  CDwarf()
  {
    GSE_Quaternion q1,q2;
    q1.CreateFromAxisAngle( 1.0, 0.0f, 0.0f, -90.0f);
    q2.CreateFromAxisAngle( 0,0,1, 30.0f);
    // first -90 over x, then 30 over z
    m_mRightHandTransform =  (q2 * q1).ToMatrix();
    // Then align handle to hand
    m_mRightHandTransform =  Math::TranslationMatrix(1.5,-4.7f,13.0) * m_mRightHandTransform;
    /////////////////////////////////////////////////////////////////
    q2.CreateFromAxisAngle( 0,0,1, -30.0f);
    // first -90 over x, then -30 over z
    m_mLeftHandTransform =  (q2 * q1).ToMatrix();
    // Then align handle to lefthand
    m_mLeftHandTransform =  Math::TranslationMatrix(-1.5,-4.7f,13.0) * m_mLeftHandTransform;
    
    m_bRightHandFull = 0;
    m_bLeftHandFull = 0;
  }
  /// Destructor.
  ~CDwarf()
  {
    
  }

public:
  /////////////////////////////////////////////////////////////////
  /// This method constructs the dwarf.
  void Construct( )
  {
    /// We load animated model dwarf_noaxe.ms3d
    GSE_MS3DModel model;
    model.Load("./data/dwarf_noaxe.ms3d");    
    m_fAnimationFPS = model.m_Animationdata.fAnimationFPS;
    GSE_GeometryNode *pDwarfGeometry  = NULL;
    m_pDwarfRootJoint = NULL;
    GSE_SceneGraph *pGraph = (GSE_SceneGraph *)GetGraph();

    /// This converts data from model into renderable and animatable form.
    GSE_MS3DObjectConverter *pMS3DConverter = 
       GSE_MS3DObjectConverter::GetInstance();
    pMS3DConverter->SetTexturePath(".");
    pMS3DConverter->Convert( model, 
			     &pGraph->GetRenderGraph(),
			     &pGraph->GetAnimationGraph(),
			     &pDwarfGeometry, 
			     &m_pDwarfRootJoint, 
			     "base",
			     TextureQuality::PERFORMANCE);

    assert(pDwarfGeometry != NULL);
    SetRenderNode( pDwarfGeometry );

    /// Retrieve root node from animation graph
    GSE_ARootNode *pARoot = (GSE_ARootNode *)(pGraph->GetAnimationGraph().SeekNodeByNameAndType( "Root", Animation::ROOT ));
    assert(pARoot != NULL);
    /// Add dwarf to animation graph.
    pARoot->AddEdge( m_pDwarfRootJoint );

    /// Here we read animation configuration for dwarf
    GSE_AnimationConfig config;
    config.ReadConfig("./data/dwarf1.model");
    
    list<string> anims = config.GetChildren("");
    list<string>::iterator it = anims.begin();
    
    for( ; it != anims.end(); it++)
    {
      unsigned int nStartMS = (unsigned int )(config.GetInt( (*it)+"|start") * 1000.0f/m_fAnimationFPS);
      unsigned int nEndMS   = (unsigned int )(config.GetInt( (*it)+"|end"  ) * 1000.0f/m_fAnimationFPS);
      int bLooping          = config.GetBoolean( (*it)+"|looping"  );
      AnimationSequence tmp(nStartMS,nEndMS,bLooping);
      m_AnimationSequences.insert( make_pair(*it,tmp) );

    }
    /// Animations need to be initialized before use, we do it here.
    pGraph->GetAnimationRenderer().SetRenderPass( GSE_AnimationRenderer::RENDER_PASS_JOINT_SETUP );
    Graph::TravelDF<Animation::NodeType,GSE_AnimationRenderer>(m_pDwarfRootJoint, &pGraph->GetAnimationRenderer());
    config.Reset();
    

    GSE_RenderGraph &rRenderGraph  = pGraph->GetRenderGraph();
    SetLightClusterNode(rRenderGraph.CreateNode<GSE_LightClusterNode>());
    /////////////////////////////////////////////////////////////////
    // Calculate bounding volumes
    GSE_GeometryNode *pGeom = (GSE_GeometryNode *)Graph::SeekNextWithType<RenderNode::NodeType>(pDwarfGeometry,RenderNode::GEOMETRY);
    assert(pGeom != NULL );
    
    GSE_Sphere boundingSphere = Geometry::CalculateBoundingSphereTight(pDwarfGeometry->GetGeometryDataWorld()->m_pVertices,
								       pDwarfGeometry->GetGeometryDataWorld()->m_iNumElements);

    GSE_OrientedBox boundingBox = Geometry::CalculateOrientedBoundingBox(pDwarfGeometry->GetGeometryDataWorld()->m_pVertices,
									 pDwarfGeometry->GetGeometryDataWorld()->m_iNumElements);
    SetLocalBoundingSphere( boundingSphere );
    SetLocalBoundingBox( boundingBox );
    
    ////////////////////
    // Create subgraph of the limbs, so breath and weapons can be attached to correct location.
    GSE_MS3DObjectConverter::GetInstance()->CreateSceneLimbsFromJoints( this, m_pDwarfRootJoint);
    m_pWeaponGrabRight = static_cast<GSE_SceneLimbNode *>(pGraph->SeekNodeByNameAndType( "rwrist", Scene::LIMB ));
    m_pWeaponGrabLeft  = static_cast<GSE_SceneLimbNode *>(pGraph->SeekNodeByNameAndType( "lwrist", Scene::LIMB ));
    m_pHead            = static_cast<GSE_SceneLimbNode *>(pGraph->SeekNodeByNameAndType( "head",   Scene::LIMB ));

    assert(m_pWeaponGrabRight != NULL);
    assert(m_pWeaponGrabLeft != NULL);
    assert(m_pHead != NULL);
    /// Create new scenegraph node for particlesystem.
    m_pBreath = GetGraph()->CreateNode<CSteamingBreath>();
    m_pBreath->Construct();

    /// Now you might ask why don't we add particlesystem to drawf, but to root node of scene graph?
    /// this is because particlesystem would move when you move the dwarf, and steam clouds don't
    /// follow you when you breath in cold weather, do they? :) 
    GetGraph()->GetRoot()->AddEdge( m_pBreath );
    
  }
  /////////////////////////////////////////////////////////////////
  /// This selects animation by name, which is defined in animation config.
  void SelectAnimation( std::string sAnimation )
  {
    map<string, AnimationSequence>::iterator it = m_AnimationSequences.find( sAnimation );
    if ( it != m_AnimationSequences.end())
    {
      GSE_SceneGraph *pGraph = (GSE_SceneGraph *)GetGraph();
      pGraph->GetAnimationRenderer().SelectAnimationSequence( (it->second).m_nStart,
							      (it->second).m_nEnd, 
							      (it->second).m_bLooping);
      pGraph->GetAnimationRenderer().SetRenderPass( GSE_AnimationRenderer::RENDER_PASS_SELECT_ANIMATION_TIME_INTERVAL );
      Graph::TravelDF<Animation::NodeType,GSE_AnimationRenderer>(m_pDwarfRootJoint, &pGraph->GetAnimationRenderer() );
    }
    
  }
  /// This is used to create breathing when dwarf is not doing anything.
  void CheckAnimationState()
  {
    if ( IsAnimationDone())
    {
      SelectAnimation("Idle");
      Breathe();
    }

  }
  /// This checks if current animation has finished.
  inline int IsAnimationDone() const
  {
    return m_pDwarfRootJoint->GetJoint()->IsAnimationFinished();
  }
  /////////////////////////////////////////////////////////////////
  /// Dwarf grabs an object to right hand.
  void GrabIntoRightHand( GSE_SceneObjectNode *pNode )
  {
    if ( !m_bRightHandFull )
    {
      m_pWeaponGrabRight->AddEdge( pNode );
      m_pWeaponRight = pNode;
      pNode->SetLocalTransform( m_mRightHandTransform);
      m_bRightHandFull = 1;
    }
  }
  /////////////////////////////////////////////////////////////////
  /// Dwarf drops an object from right hand, if there was any.
  GSE_SceneObjectNode * DropFromRightHand()
  {
    if ( m_bRightHandFull )
    {
      m_pWeaponGrabRight->DeleteEdgeTo( m_pWeaponRight );
      GSE_SceneObjectNode *pTmp = m_pWeaponRight;
      m_pWeaponRight = NULL;
      m_bRightHandFull = 0;
      return pTmp;
    } 
    return NULL;

  }
  /////////////////////////////////////////////////////////////////
  /// Dwarf grabs an object to left hand.
  void GrabIntoLeftHand ( GSE_SceneObjectNode *pNode )
  {
    if ( !m_bLeftHandFull)
    {
      m_pWeaponGrabLeft->AddEdge( pNode );
      m_pWeaponLeft = pNode;
      pNode->SetLocalTransform( m_mLeftHandTransform);
      m_bLeftHandFull = 1;     
    }
  }
  /////////////////////////////////////////////////////////////////
  /// This attaches and object to some relative location from right hand.
  void AttachToHand( GSE_SceneNode *pNode )
  {
    m_pWeaponGrabRight->AddEdge(pNode);
  }
  /////////////////////////////////////////////////////////////////
  /// This method creates breath.
  void Breathe()
  {

    ////////////////////
    // Here we transform the direction and position according to the transforms
    GSE_Vector3 vVelocity(0,-1,-1);
    GSE_Vector3 vMouth( 0.0f, 1.0f,  -4.5f );

    GSE_Matrix4x4f mMouthTransform = Math::TranslationMatrix( vMouth );
    GSE_Vector3 vHead( m_pHead->GetWorldTransform().At(0,3),
		       m_pHead->GetWorldTransform().At(1,3),
		       m_pHead->GetWorldTransform().At(2,3) );
    mMouthTransform = m_pHead->GetWorldTransform() * mMouthTransform;
    vMouth[GSE_Vector3::X] = mMouthTransform.At(0,3);
    vMouth[GSE_Vector3::Y] = mMouthTransform.At(1,3);
    vMouth[GSE_Vector3::Z] = mMouthTransform.At(2,3);
    vVelocity = Rotate(vVelocity, mMouthTransform);
    vVelocity.ToUnit();
    m_pBreath->Breathe( vMouth, vVelocity);
  }

  /// Returns particle system.
  CSteamingBreath *GetBreath()
  {
    return m_pBreath;
  }
};
///////////////////////////////////////////////////////////////////
/// Favourite dwarven weapon.
class CAxe : public GSE_SceneObjectNode 
{
  friend class GSE_SceneGraph;
protected:

  CAxe() {}
  ~CAxe() {}

public:
  /////////////////////////////////////////////////////////////////
  /// Duplicates this node, several axes are vewy vewy cool thing.
  void Duplicate( CAxe &axe)
  {
    GSE_RenderGraph &rRenderGraph  = ((GSE_SceneGraph *)GetGraph())->GetRenderGraph();
    SetLightClusterNode(rRenderGraph.CreateNode<GSE_LightClusterNode>());
    SetRenderNode(axe.GetRenderNode());
    SetLocalBoundingSphere( axe.GetLocalBoundingSphere() );
    SetLocalBoundingBox( axe.GetLocalBoundingBox() );
  }
  /////////////////////////////////////////////////////////////////
  /// This method creates the axe.
  void Construct()
  {
    GSE_MS3DModel model;
    model.Load("./data/axe.ms3d");    
    GSE_GeometryNode *pAxeGeometry  = NULL;
    GSE_SceneGraph *pGraph = (GSE_SceneGraph *)GetGraph();
    GSE_JointNode *pJoint = NULL;
    /////////////////////////////////////////////////////////////////
    // Convert model into GSE rendergraph
    GSE_MS3DObjectConverter::GetInstance()->Convert( model, 
						     &pGraph->GetRenderGraph(),
						     NULL,
						     &pAxeGeometry, 
						     &pJoint, 
						     "",
						     TextureQuality::QUALITY);
    assert( pAxeGeometry != NULL );
    SetRenderNode( pAxeGeometry );
    /////////////////////////////////////////////////////////////////
    // Enable lighting
    GSE_RenderGraph &rRenderGraph  = pGraph->GetRenderGraph();
    SetLightClusterNode(rRenderGraph.CreateNode<GSE_LightClusterNode>());
    /////////////////////////////////////////////////////////////////
    // Calculate bounding volumes
    GSE_GeometryNode *pGeom = (GSE_GeometryNode *)Graph::SeekNextWithType<RenderNode::NodeType>(pAxeGeometry,RenderNode::GEOMETRY);
    assert(pGeom != NULL );
    GSE_Sphere boundingSphere = Geometry::CalculateBoundingSphereTight(pGeom->GetGeometryData()->m_pVertices, 
								       pGeom->GetGeometryData()->m_iNumElements);
    GSE_OrientedBox boundingBox = Geometry::CalculateOrientedBoundingBox( pGeom->GetGeometryData()->m_pVertices, 
								  pGeom->GetGeometryData()->m_iNumElements );
    SetLocalBoundingSphere( boundingSphere );
    SetLocalBoundingBox( boundingBox );

  }
  
};
/////////////////////////////////////////////////////////////////
/// This scenegraph node is part of the ruins.
class CRuinsFloor : public GSE_SceneObjectNode 
{
  friend class GSE_SceneGraph;
protected:
   CRuinsFloor() {}
  ~CRuinsFloor() {}
public:
  /// This method constructs the ruins floor.
  void Construct()
  {
    GSE_SceneGraph *pGraph = (GSE_SceneGraph *)GetGraph();
    GSE_GeometryNode *pGeometry = NULL;
    /// Load 3DS model
    GSE_3dsConverter::GetInstance()->Convert( "./data/ruins33.3DS", "floor",
					      &pGraph->GetRenderGraph(),
					      &pGeometry );
    assert(pGeometry != NULL );
    SetRenderNode(pGeometry);
    /// Calculate bounding volumes.
    GSE_Sphere sphere   = Geometry::CalculateBoundingSphereTight( pGeometry->GetGeometryData()->m_pVertices,
								pGeometry->GetGeometryData()->m_iNumElements );
    GSE_OrientedBox box = Geometry::CalculateOrientedBoundingBox( pGeometry->GetGeometryData()->m_pVertices,
								  pGeometry->GetGeometryData()->m_iNumElements );
    
    SetLocalBoundingSphere( sphere );
    SetLocalBoundingBox( box );
  }
};
/////////////////////////////////////////////////////////////////
/// This scenegraph node is part of the ruins.
class CRuinsRubble : public GSE_SceneObjectNode 
{
  friend class GSE_SceneGraph;
protected:
   CRuinsRubble() {}
  ~CRuinsRubble() {}
public:
  /// This method constructs the rubble of the ruins.
  void Construct()
  {
    GSE_SceneGraph *pGraph = (GSE_SceneGraph *)GetGraph();
    GSE_GeometryNode *pGeometry = NULL;
    /// Load 3DS model.
    GSE_3dsConverter::GetInstance()->Convert( "./data/ruins33.3DS", "Rubble",
					      &pGraph->GetRenderGraph(),
					      &pGeometry );
    assert(pGeometry != NULL );
    SetRenderNode(pGeometry);
    /// Calculate bounding volumes.
    GSE_Sphere sphere   = Geometry::CalculateBoundingSphereTight( pGeometry->GetGeometryData()->m_pVertices,
								pGeometry->GetGeometryData()->m_iNumElements );
    GSE_OrientedBox box = Geometry::CalculateOrientedBoundingBox( pGeometry->GetGeometryData()->m_pVertices,
								  pGeometry->GetGeometryData()->m_iNumElements );
    
    SetLocalBoundingSphere( sphere );
    SetLocalBoundingBox( box );
  }
};
/////////////////////////////////////////////////////////////////
/// This scenegraph node is part of the ruins.
class CRuinsWall1 : public GSE_SceneObjectNode 
{
  friend class GSE_SceneGraph;
protected:
   CRuinsWall1() {}
  ~CRuinsWall1() {}
public:
  void Construct()
  {
    GSE_SceneGraph *pGraph = (GSE_SceneGraph *)GetGraph();
    GSE_GeometryNode *pGeometry = NULL;
    /// Load 3DS Model
    GSE_3dsConverter::GetInstance()->Convert( "./data/ruins33.3DS", "wall1",
					      &pGraph->GetRenderGraph(),
					      &pGeometry );
    assert(pGeometry != NULL );
    SetRenderNode(pGeometry);
    /// Calculate bounding volumes.
    GSE_Sphere sphere   = Geometry::CalculateBoundingSphereTight( pGeometry->GetGeometryData()->m_pVertices,
								pGeometry->GetGeometryData()->m_iNumElements );
    GSE_OrientedBox box = Geometry::CalculateOrientedBoundingBox( pGeometry->GetGeometryData()->m_pVertices,
								  pGeometry->GetGeometryData()->m_iNumElements );
    
    SetLocalBoundingSphere( sphere );
    SetLocalBoundingBox( box );
  }
};
/////////////////////////////////////////////////////////////////
/// This scenegraph node is part of the ruins.
class CRuinsWall2 : public GSE_SceneObjectNode 
{
  friend class GSE_SceneGraph;
protected:
   CRuinsWall2() {}
  ~CRuinsWall2() {}
public:
  void Construct()
  {
    GSE_SceneGraph *pGraph = (GSE_SceneGraph *)GetGraph();
    GSE_GeometryNode *pGeometry = NULL;
    /// Load 3DS model
    GSE_3dsConverter::GetInstance()->Convert( "./data/ruins33.3DS", "wall2",
					      &pGraph->GetRenderGraph(),
					      &pGeometry );
    assert(pGeometry != NULL );
    SetRenderNode(pGeometry);
    /// Calculate bounding volumes.
    GSE_Sphere sphere   = Geometry::CalculateBoundingSphereTight( pGeometry->GetGeometryData()->m_pVertices,
								pGeometry->GetGeometryData()->m_iNumElements );
    GSE_OrientedBox box = Geometry::CalculateOrientedBoundingBox( pGeometry->GetGeometryData()->m_pVertices,
								  pGeometry->GetGeometryData()->m_iNumElements );
    
    SetLocalBoundingSphere( sphere );
    SetLocalBoundingBox( box );
  }
};
/////////////////////////////////////////////////////////////////
/// This scenegraph node connects other parts of the ruins.
class CRuins : public GSE_SceneNode 
{
  friend class GSE_SceneGraph;
protected:
   CRuins() {}
  ~CRuins() {}
public:
  ////////////////////
  /// The construction method.
  void Construct()
  {
    GSE_SceneGraph *pGraph = (GSE_SceneGraph *)GetGraph();
    /// Create nodes for floor, rubble, and walls.
    CRuinsFloor *pFloor = pGraph->CreateNode<CRuinsFloor>();
    CRuinsRubble *pRubble = pGraph->CreateNode<CRuinsRubble>();
    CRuinsWall1 *pWall1 = pGraph->CreateNode<CRuinsWall1>();
    CRuinsWall2 *pWall2 = pGraph->CreateNode<CRuinsWall2>();
    /// Construct nodes.
    pFloor->Construct();
    pRubble->Construct();
    pWall1->Construct();
    pWall2->Construct();
    /// Add edges to those nodes.
    AddEdge(pFloor);
    AddEdge(pRubble);
    AddEdge(pWall1);
    AddEdge(pWall2);
  }
};
/////////////////////////////////////////////////////////////////
/// This scenegraph node creates the ground using Octree.
class CGround : public GSE_SceneOctreeObjectNode 
{
  friend class GSE_SceneGraph;
protected:
   CGround() {}
  ~CGround() {}
public:
  ////////////////////
  /// The construction method.
  void Construct()
  {
    GSE_GeometryOctree *pOctree = NULL;
    GSE_GeometryNode *pGeometryNode = NULL;
    GSE_TerrainConverter *pTerrainConv = GSE_TerrainConverter::GetInstance();

    GSE_SceneGraph *pSceneGraph = GetGraph();
    GSE_Vector3 vCenter(0,-50,0);
    /// Load terrain.
    pTerrainConv->CreateTerrainOctreeFromFile( "./data/terraintest.png",
					       *pSceneGraph,
					       &pOctree, &pGeometryNode,
					       21.0f, 125.0f, 3, 124, vCenter);

    assert(pOctree != NULL );
    SetGeometryOctree(pOctree);

    GSE_RenderGraph &rRenderGraph = static_cast<GSE_SceneGraph *>(GetGraph())->GetRenderGraph();
    /// Load ground texture.
    GSE_Texture *pGroundTexture = 
           GSE_OglTextureMgr::GetInstance()->CreateTextureFromImage( "./data/snow.jpg", 0, 
								     TextureType::GSE_TEX2D, 
								     TextureQuality::QUALITY);
    GSE_TextureNode *pGroundTextureNode = rRenderGraph.CreateNode<GSE_TextureNode>();

    pGroundTextureNode->SetTexture(pGroundTexture);
    pGroundTextureNode->AddEdge(pGeometryNode);
    SetRenderNode( pGroundTextureNode );
    SetGeometryNode( pGeometryNode );

    /// Calculate bounding volumes.
    GSE_GeometryData *pGeometry = pGeometryNode->GetGeometryData();
    GSE_Sphere sphere = Geometry::CalculateBoundingSphereTight(pGeometry->m_pVertices,
							       pGeometry->m_iNumElements );
    SetLocalBoundingSphere(sphere);
    GSE_OrientedBox obBox;
    obBox.SetOrientation( GSE_Vector3::GetWorldY(), GSE_Vector3::GetWorldZ(), 
			  GSE_Vector3::GetWorldX());
    obBox.SetWidth(  pOctree->GetWidth());
    obBox.SetHeight( pOctree->GetHeight());
    obBox.SetLength( pOctree->GetLength());
    obBox.SetPosition( pOctree->GetPosition());
    
    SetLocalBoundingBox( obBox );
  }
};
/////////////////////////////////////////////////////////////////
#define SKYBOX_VERTEX_COUNT 24
#define QUAD_VERTEX_COUNT 4
#define QUADS_IN_SKYBOX 6
#define SKYBOX_SIZE 2.0f
/////////////////////////////////////////////////////////////////
/// This node creates a skybox, which is a cube and rendered in special
/// way to simulate sky.
class CSkybox: public GSE_SceneObjectNode
{
  friend class GSE_SceneGraph;
protected:
   CSkybox() {}
  ~CSkybox() {}

  void CreateBoxWall( std::string sTextureFilename, GSE_GeometryNode *pGeomNode, 
		      unsigned int nIndexOne, unsigned int nIndexTwo, 
		      unsigned int nIndexThree, unsigned int nIndexFour)
  {
    GSE_TextureNode *pSkyboxTexNode = NULL;
    GSE_Texture *pSkyboxTexture     = NULL;
    GSE_IndexBuffer *pIndexBuffer  = NULL;
    GSE_IndexBufferNode *pIndexBufferNode  = NULL;
    ////////////////////
    GSE_RenderGraph &rRenderGraph  = GetGraph()->GetRenderGraph();
    pSkyboxTexture = GSE_OglTextureMgr::GetInstance()->CreateTextureFromImage( const_cast<char *>(sTextureFilename.c_str()),0,TextureType::GSE_TEX2D, TextureQuality::QUALITY);
    assert(pSkyboxTexture != NULL );

    pSkyboxTexNode = rRenderGraph.CreateNode<GSE_TextureNode>();
    pSkyboxTexNode->SetTexture( pSkyboxTexture );
    pSkyboxTexNode->Add(TextureFilter::T_WRAP_CLAMP_TO_EDGE);
    pSkyboxTexNode->Add(TextureFilter::S_WRAP_CLAMP_TO_EDGE);
    pIndexBuffer = GSE_IndexBufferMgr::GetInstance()->CreateIndexBuffer();
    pIndexBuffer->m_nVerticesInPrimitive = QUAD_VERTEX_COUNT; 
    pIndexBuffer->Init( QUAD_VERTEX_COUNT );
    pIndexBuffer->m_pIndices[0] = nIndexOne;
    pIndexBuffer->m_pIndices[1] = nIndexTwo;
    pIndexBuffer->m_pIndices[2] = nIndexThree;
    pIndexBuffer->m_pIndices[3] = nIndexFour;
    pIndexBufferNode = rRenderGraph.CreateNode<GSE_IndexBufferNode>();
    pIndexBufferNode->SetIndexBuffer( pIndexBuffer );
    pGeomNode->AddEdge( pSkyboxTexNode );
    pSkyboxTexNode->AddEdge( pIndexBufferNode ); 
    
  }
public:

  
  void Construct()
  {

    GSE_RenderGraph &rRenderGraph  = GetGraph()->GetRenderGraph();
    GSE_GeometryNode *pGeomNode = rRenderGraph.CreateNode<GSE_GeometryNode>();
    pGeomNode->SetGeometryData(GSE_GeometryDataMgr::GetInstance()->CreateGeometryData());
    pGeomNode->SetGeometryDataWorld(pGeomNode->GetGeometryData());
    ////////////////////
    // Six quads with vertices that have different texcoords
    ////////////////////
    pGeomNode->GetGeometryData()->SetVertexCount( SKYBOX_VERTEX_COUNT );
    pGeomNode->GetGeometryData()->InitVertices();
    pGeomNode->GetGeometryData()->InitTexCoords(0);
    ////////////////////
    // First quad
    ////////////////////
    GSE_Vector3 vVertex(-SKYBOX_SIZE,SKYBOX_SIZE,-SKYBOX_SIZE);
    pGeomNode->GetGeometryData()->SetVertex(0,vVertex);
    pGeomNode->GetGeometryData()->SetTexCoords(0,0, 1.0f,0.0f);
    vVertex[GSE_Vector3::Y] = -SKYBOX_SIZE;
    pGeomNode->GetGeometryData()->SetVertex(1,vVertex);
    pGeomNode->GetGeometryData()->SetTexCoords(0,1, 1.0f,1.0f);
    vVertex[GSE_Vector3::X] = SKYBOX_SIZE;
    pGeomNode->GetGeometryData()->SetVertex(2,vVertex);
    pGeomNode->GetGeometryData()->SetTexCoords(0,2, 0.0f,1.0f);
    vVertex[GSE_Vector3::Y] = SKYBOX_SIZE;
    pGeomNode->GetGeometryData()->SetVertex(3,vVertex);
    pGeomNode->GetGeometryData()->SetTexCoords(0,3, 0.0f,0.0f);
    ////////////////////
    // second quad
    ////////////////////
    vVertex =GSE_Vector3(SKYBOX_SIZE,SKYBOX_SIZE,-SKYBOX_SIZE);
    pGeomNode->GetGeometryData()->SetVertex(4,vVertex);
    pGeomNode->GetGeometryData()->SetTexCoords(0,4, 1.0f,0.0f);
    vVertex[GSE_Vector3::Y] = -SKYBOX_SIZE;
    pGeomNode->GetGeometryData()->SetVertex(5,vVertex);
    pGeomNode->GetGeometryData()->SetTexCoords(0,5, 1.0f,1.0f);
    vVertex[GSE_Vector3::Z] = SKYBOX_SIZE;
    pGeomNode->GetGeometryData()->SetVertex(6,vVertex);
    pGeomNode->GetGeometryData()->SetTexCoords(0,6, 0.0f,1.0f);
    vVertex[GSE_Vector3::Y] = SKYBOX_SIZE;
    pGeomNode->GetGeometryData()->SetVertex(7,vVertex);
    pGeomNode->GetGeometryData()->SetTexCoords(0,7, 0.0f,0.0f);
    ////////////////////
    // third quad
    ////////////////////
    vVertex =GSE_Vector3(SKYBOX_SIZE,SKYBOX_SIZE,SKYBOX_SIZE);
    pGeomNode->GetGeometryData()->SetVertex(8,vVertex);
    pGeomNode->GetGeometryData()->SetTexCoords(0,8, 1.0f,0.0f);
    vVertex[GSE_Vector3::Y] = -SKYBOX_SIZE;
    pGeomNode->GetGeometryData()->SetVertex(9,vVertex);
    pGeomNode->GetGeometryData()->SetTexCoords(0,9, 1.0f,1.0f);
    vVertex[GSE_Vector3::X] = -SKYBOX_SIZE;
    pGeomNode->GetGeometryData()->SetVertex(10,vVertex);
    pGeomNode->GetGeometryData()->SetTexCoords(0,10, 0.0f,1.0f);
    vVertex[GSE_Vector3::Y] = SKYBOX_SIZE;
    pGeomNode->GetGeometryData()->SetVertex(11,vVertex);
    pGeomNode->GetGeometryData()->SetTexCoords(0,11, 0.0f,0.0f);
    ////////////////////
    // fourth quad
    ////////////////////
    vVertex =GSE_Vector3(-SKYBOX_SIZE,SKYBOX_SIZE,SKYBOX_SIZE);
    pGeomNode->GetGeometryData()->SetVertex(12,vVertex);
    pGeomNode->GetGeometryData()->SetTexCoords(0,12, 1.0f,0.0f);
    vVertex[GSE_Vector3::Y] = -SKYBOX_SIZE;
    pGeomNode->GetGeometryData()->SetVertex(13,vVertex);
    pGeomNode->GetGeometryData()->SetTexCoords(0,13, 1.0f,1.0f);
    vVertex[GSE_Vector3::Z] = -SKYBOX_SIZE;
    pGeomNode->GetGeometryData()->SetVertex(14,vVertex);
    pGeomNode->GetGeometryData()->SetTexCoords(0,14, 0.0f,1.0f);
    vVertex[GSE_Vector3::Y] = SKYBOX_SIZE;
    pGeomNode->GetGeometryData()->SetVertex(15,vVertex);
    pGeomNode->GetGeometryData()->SetTexCoords(0,15, 0.0f,0.0f);
    ////////////////////
    // fifth quad
    ////////////////////
    vVertex =GSE_Vector3(-SKYBOX_SIZE,SKYBOX_SIZE,SKYBOX_SIZE);
    pGeomNode->GetGeometryData()->SetVertex(16,vVertex);
    pGeomNode->GetGeometryData()->SetTexCoords(0,16, 1.0f,1.0f);
    vVertex[GSE_Vector3::Z] = -SKYBOX_SIZE;
    pGeomNode->GetGeometryData()->SetVertex(17,vVertex);
    pGeomNode->GetGeometryData()->SetTexCoords(0,17, 0.0f,1.0f);
    vVertex[GSE_Vector3::X]= SKYBOX_SIZE;
    pGeomNode->GetGeometryData()->SetVertex(18,vVertex);
    pGeomNode->GetGeometryData()->SetTexCoords(0,18, 0.0f,0.0f);
    vVertex[GSE_Vector3::Z]= SKYBOX_SIZE;
    pGeomNode->GetGeometryData()->SetVertex(19,vVertex);
    pGeomNode->GetGeometryData()->SetTexCoords(0,19, 1.0f,0.0f);
    
    ////////////////////
    // sixth quad
    ////////////////////
    vVertex =GSE_Vector3(-SKYBOX_SIZE,-SKYBOX_SIZE,-SKYBOX_SIZE);
    pGeomNode->GetGeometryData()->SetVertex(20,vVertex);
    pGeomNode->GetGeometryData()->SetTexCoords(0,20, 1.0f,0.0f);

    vVertex[GSE_Vector3::Z] = SKYBOX_SIZE;
    pGeomNode->GetGeometryData()->SetVertex(21,vVertex);
    pGeomNode->GetGeometryData()->SetTexCoords(0,21, 1.0f,1.0f);

    vVertex[GSE_Vector3::X] = SKYBOX_SIZE;
    pGeomNode->GetGeometryData()->SetVertex(22,vVertex);
    pGeomNode->GetGeometryData()->SetTexCoords(0,22, 0.0f,1.0f);
    
    vVertex[GSE_Vector3::Z] = -SKYBOX_SIZE;
    pGeomNode->GetGeometryData()->SetVertex(23,vVertex);
    pGeomNode->GetGeometryData()->SetTexCoords(0,23, 0.0f,0.0f);    
    
    GSE_DepthBufferStateNode *pDepthBufState = rRenderGraph.CreateNode<GSE_DepthBufferStateNode>();
    cerr << "SKYBOX DEPTH BUFFER STATE NODE : " << pDepthBufState << endl;
    pDepthBufState->SetDepthTest(0);
    pDepthBufState->SetDepthBufferWriting(0);
    
    //assert(pDepthBufState->IsDepthTest() == 0 );
    //assert(pDepthBufState->IsDepthBufferWriting() == 0);
    
    GSE_ColorNode *pColor = rRenderGraph.CreateNode<GSE_ColorNode>();
    pColor->Set( 1,1,1,1);
    GSE_ClearBufferNode *pClearBufNode = rRenderGraph.CreateNode<GSE_ClearBufferNode>();
    cerr << "SKYBOX CLEAR BUFFER NODE " << pClearBufNode << endl;
    pClearBufNode->SetClearColorBuffer(0);
    pClearBufNode->SetClearDepthBuffer(1);
    
    SetRenderNode( pClearBufNode );          // Clear depth buffer only
    pClearBufNode->AddEdge( pDepthBufState ); // Set writing to depth buffer off
    pDepthBufState->AddEdge( pColor );
    pColor->AddEdge( pGeomNode);  // Draw the skybox
    
    CreateBoxWall( "./data/skybox/arcnbk.bmp",pGeomNode, 0,1,2,3 );
    CreateBoxWall( "./data/skybox/arcnlf.bmp",pGeomNode, 4,5,6,7 );
    CreateBoxWall( "./data/skybox/arcnft.bmp",pGeomNode, 8,9,10,11 );
    CreateBoxWall( "./data/skybox/arcnrt.bmp",pGeomNode, 12,13,14,15 );
    CreateBoxWall( "./data/skybox/arcnup.bmp",pGeomNode, 16,17,18,19 );
    CreateBoxWall( "./data/skybox/arcndn.bmp",pGeomNode, 20,21,22,23 );
  }
};
/////////////////////////////////////////////////////////////////
#undef SKYBOX_VERTEX_COUNT 
#undef QUAD_VERTEX_COUNT 
#undef QUADS_IN_SKYBOX     
///////////////////////////////////////////////////////////////////
#define SENSITIVITY 0.5f
#define MOVEMENT_SPEED 123.0f
/////////////////////////////////////////////////////////////////
int main()
{
  bool bLoop = 1;   // is main loop active
  int bMousePressed = 0; //used in input handling
  unsigned int bEventsFound = 0; // used in input handling
  GSE_Vector2 vOld,vNew; // mouse positions.


  GSE_OglScreenParams params;
  params.m_iDepthBufferSize = 24;

  /////////////////////////////////////////////////////////////////
  /// Initialize sound system
  OpenALSoundSystem alSoundSystem;
  alSoundSystem.LoadSound( 0, "./data/explosion.wav" );
  OpenALSoundSource *pExplosion = alSoundSystem.CreateSource(0);
  /////////////////////////////////////////////////////////////////
  /// Screen must be opened first.
  GSE_Screen *pScreen = GSE_Screen::GetScreen( params);
  if ( pScreen == NULL )
  {
    cerr << "Couldn't open screen" << endl;
    exit(1);
  }
  /////////////////////////////////////////////////////////////////
  /// Create scene graph and objects in our scene.
  GSE_SceneGraph *pSceneGraph = new GSE_SceneGraph();
  CDwarf *pDwarf  = pSceneGraph->CreateNode<CDwarf>();
  CAxe *pAxe      = pSceneGraph->CreateNode<CAxe>();
  CAxe *pAnotherAxe = pSceneGraph->CreateNode<CAxe>();
  CRuins *pRuins = pSceneGraph->CreateNode<CRuins>();
  CGround *pGround = pSceneGraph->CreateNode<CGround>();
  CSkybox *pSkybox = pSceneGraph->CreateNode<CSkybox>();

  
  pDwarf->Construct();
  pAxe->Construct();
  pAnotherAxe->Duplicate( *pAxe );
  pRuins->Construct();
  pGround->Construct();
  pSkybox->Construct();
  /// Skybox is set in different manner.
  pSceneGraph->SetSkybox( pSkybox );
  GSE_RenderGraph *pRenderGraph = &pSceneGraph->GetRenderGraph();
  /////////////////////////////////////////////////////////////////
  // Camera is created here.
  GSE_Camera *pCamera = GSE_CameraMgr::GetInstance()->CreateCamera();
  GSE_SceneCameraNode *pCameraNode = pSceneGraph->CreateNode<GSE_SceneCameraNode>();
  /// Camera settings
  pCameraNode->SetCamera(pCamera);
  pCamera->SetViewport(0, 0, 640, 480);
  pCamera->SetFieldOfView(43.5);
  pCamera->SetNearClipping(0.01f);
  pCamera->SetFarClipping(1000.0f);
  pCamera->SetPosition( GSE_Vector3(0.0, 0.0, 74.0));
  pCamera->UpdateProjection();
  pCamera->UpdateView();
  GSE_Matrix4x4f matrix = Math::TranslationMatrix(20,0,0) * Math::RotationMatrix(0,DEG2RAD(90.0f),0);
  pCameraNode->SetLocalTransform( matrix );
  
  /////////////////////////////////////////////////////////////////
  // Clear color & depth buffer
  GSE_ClearBufferNode *pClear = pRenderGraph->CreateNode<GSE_ClearBufferNode>();
  pClear->SetClearColorBuffer(0);
  pClear->SetClearDepthBuffer(0);
  pClear->ClearColor().Set(1.0, 0.0, 0.0, 1.0);
  /////////////////////////////////////////////////////////////////
  // Depth test on
  GSE_DepthBufferStateNode *pDepthBufferState = pRenderGraph->CreateNode<GSE_DepthBufferStateNode>();
  pDepthBufferState->SetDepthTest(1);
  pDepthBufferState->SetDepthBufferWriting(1);
  
  /////////////////////////////////////////////////////////////////
  /// Lighting is created here.
  GSE_SceneLightNode *pLightNode = pSceneGraph->CreateNode<GSE_SceneLightNode>();
  pLightNode->SetLight(GSE_LightMgr::GetInstance()->GetLightFromFile( "./data/Lights.cfg", "Yellow"));
  assert( pLightNode->GetLight() != NULL );
  /////////////////////////////////////////////////////////////////
  // Backface Culling on
  GSE_CullingStateNode *pCulling = pRenderGraph->CreateNode<GSE_CullingStateNode>();
  pCulling->SetBackfaceCulling(1);
  pCulling->SetFaceCulling(1);
  /////////////////////////////////////////////////////////////////  
  // Construct the render settings for dwarf and weapons.
  GSE_SceneRenderSettingsChainNode *pRenderSettings = pSceneGraph->CreateNode<GSE_SceneRenderSettingsChainNode>();  
  pRenderSettings->AddLink( pClear );    
  pRenderSettings->AddLink( pCulling );
  pRenderSettings->AddLink( pDepthBufferState );
  /////////////////////////////////////////////////////////////////
  /// Ground is not drawn with rendersettings
  pSceneGraph->GetRoot()->AddEdge( pGround );
  /////////////////////////////////////////////////////////////////
  /// Dwarf is rendered using the defined rendersettings.
  pSceneGraph->GetRoot()->AddEdge( pRenderSettings );
  pRenderSettings->AddEdge( pDwarf );
  pDwarf->GrabIntoRightHand( pAxe );
  pDwarf->GrabIntoLeftHand(pAnotherAxe);
  /////////////////////////////////////////////////////////////////
  /// ruins are not drawn with rendersettings.
  pSceneGraph->GetRoot()->AddEdge( pRuins );
  /////////////////////////////////////////////////////////////////
  /// Here we create new opengl renderer.
  GSE_OglRenderer *pRenderer = new GSE_OglRenderer();
  /// event is for input
  SDL_Event event; 
  Core::GSE_FpsCounter fpsCounter;
  fpsCounter.SetStartTime( SDL_GetTicks());
  
  GSE_RenderRoot *pRoot = pSceneGraph->GetRenderRoot();
  /// Initial animation is idle
  pDwarf->SelectAnimation("Idle");

  unsigned int nTimeFormer = SDL_GetTicks();
  unsigned int nPassedTime = 0;
  unsigned int nTimeLatter = 0;
  /// Put dwarf in correct place
  GSE_Matrix4x4f transform = TranslationMatrix(0,10,-30) * RotationMatrix(0,DEG2RAD(180),0);
  pDwarf->SetLocalTransform( transform );
  /////////////////////////////////////////////////////////////////
  /// Set position of explosion sound.
  GSE_Vector3 vPos(0, 10, -30);
  pExplosion->SetPosition(vPos);
  pExplosion->SetGain( 4.0f );
  /////////////////////////////////////////////////////////////////
  /// Dwarf will be lit, create illumination set and attach it to dwarf and weapons.
  GSE_IlluminationSet *pSet = GSE_IlluminationSetMgr::GetInstance()->CreateIlluminationSet("BasicLighting");
  pSet->AddLight( pLightNode->GetLight() );
  matrix = Math::TranslationMatrix(-10,-13,0);
  pDwarf->SetIlluminationSet( pSet );
  pAxe->SetIlluminationSet( pSet );
  pAnotherAxe->SetIlluminationSet( pSet );

  pDwarf->AttachToHand(pLightNode);
  Core::GSE_Timer movementTimer;
  movementTimer.SetStartTime( SDL_GetTicks());

  /////////////////////////////////////////////////////////////////
  /// prepare scene graph
  pSceneGraph->UpdateTransforms();
  pSceneGraph->UpdateBoundingVolumes();
  /////////////////////////////////////////////////////////////////
  // Hide & grab mouse, center it, and speed up keyrepeats
  SDL_WM_GrabInput(SDL_GRAB_OFF);
  SDL_ShowCursor(SDL_DISABLE);
  SDL_WarpMouse( 320,480 );
  SDL_EnableKeyRepeat(1, 1);
  /////////////////////////////////////////////////////////////////
  // remove all existing events from the queue
  while ( SDL_PollEvent(&event ));

  /////////////////////////////////////////////////////////////////
  /// The main loop.
  while ( bLoop )
  {
    /// Check for input.
    while ( SDL_PollEvent(&event ))
    {
      if ( bEventsFound == 0)
      {
	movementTimer.SetCurrentTime(SDL_GetTicks());
	bEventsFound = 1;
      }
      switch(event.type)
      {
	/// When key is pressed.
      case SDL_KEYDOWN:
	/// Escape exits.
	if ( event.key.keysym.sym == SDLK_ESCAPE)	bLoop = 0;

	/// Left arrow moves camera.
	if ( event.key.keysym.sym == SDLK_LEFT) 
	{

          pCamera->Move( pCamera->GetRightVector() * -MOVEMENT_SPEED  * movementTimer.GetPassedTimeMS()*0.001f);
	  
	}
	/// Right arrow moves camera.
	if ( event.key.keysym.sym == SDLK_RIGHT)
	{          
	  pCamera->Move(pCamera->GetRightVector() * MOVEMENT_SPEED  * movementTimer.GetPassedTimeMS()*0.001f);

	}
	/// Return makes dwarf to grab weapon into its hand.
	if ( event.key.keysym.sym == SDLK_RETURN) 
	{          
	  pDwarf->GrabIntoRightHand( pAxe );
	}
	/// Backspace makes dwarf to drop weapon from its hand.
	if ( event.key.keysym.sym == SDLK_BACKSPACE) 
	{          
	  pDwarf->DropFromRightHand( );
	}
	/// Up arrow moves camera forward.
	if ( event.key.keysym.sym == SDLK_UP) 
	{
	  pCamera->Move( pCamera->GetForwardVector() * 
			 MOVEMENT_SPEED * 
			 movementTimer.GetPassedTimeMS()*0.001f);

	}
	/// PageUp moves camera up.
	if ( event.key.keysym.sym == SDLK_PAGEUP )
	{
	  pCamera->Move( pCamera->GetUpVector() * 
			 MOVEMENT_SPEED * 
			 movementTimer.GetPassedTimeMS()*0.001f);
	}
	/// PageDown moves camera down.
	if ( event.key.keysym.sym == SDLK_PAGEDOWN )
	{
	  pCamera->Move( -pCamera->GetUpVector() * 
			 MOVEMENT_SPEED * 
			 movementTimer.GetPassedTimeMS()*0.001f);
	}
	/// Down arrow moves camera backward.
	if ( event.key.keysym.sym == SDLK_DOWN) 
	{
	  pCamera->Move( -pCamera->GetForwardVector() * 
			 MOVEMENT_SPEED * 
			 movementTimer.GetPassedTimeMS()*0.001f);

	}
	/// Space makes dwarf walk.
	if ( event.key.keysym.sym == SDLK_SPACE )
	{
	  pDwarf->SelectAnimation("Walk");
	}
	/// Right shift makes dwarf hit with its weapon,
	/// and breathe. Also explosion is played.
	if ( event.key.keysym.sym == SDLK_RSHIFT )
	{
	  pDwarf->SelectAnimation("Attack1");
	  pDwarf->Breathe();
	  pExplosion->Play();
	}
	/// Update position of listener.
	alSoundSystem.SetListener( *pCamera, pCamera->GetPosition());
	break;
	/// When mouse is moved.
      case SDL_MOUSEMOTION:
	/// If mouse button is pressed, then use virtual trackball.
	if ( bMousePressed )
	{
	  
	  vNew[GSE_Vector2::X] = event.button.x;
	  vNew[GSE_Vector2::Y] = 480.0f - event.button.y;

	  GSE_Vector3 vPos;
	  GSE_Vector3 vStart,vEnd;
	  

	  vStart = pCamera->WindowCoordinatesToWorld(vNew[GSE_Vector2::X],
						     vNew[GSE_Vector2::Y],
						     0.0f); 
	  vEnd = pCamera->WindowCoordinatesToWorld(vNew[GSE_Vector2::X],
						   vNew[GSE_Vector2::Y],
						   1.0f); 

	  pCamera->VirtualTrackball( vPos, vOld, vNew );
	  vOld = vNew;
	  
	} 
	else
	{
	  /// Rotate camera.
	  pCamera->RotateAroundY( (320-event.motion.x)*SENSITIVITY );
	  pCamera->RotateAroundRight( (240-event.motion.y)*SENSITIVITY );
	  SDL_WarpMouse( 320,240 );
	  /// Update position of listener.
	  alSoundSystem.SetListener( *pCamera, pCamera->GetPosition());
	}
	
	break;
	/// When mouse button is pressed.
      case SDL_MOUSEBUTTONDOWN:
	{
	  switch( event.button.button )
	  {
	    // Moves camera forward.
	  case SDL_BUTTON_WHEELUP:
	    cerr << "wheel button up" << endl;
	    pCamera->Move( pCamera->GetForwardVector()*5.0f);
	    break;
	    // Moves camera backward.
	  case SDL_BUTTON_WHEELDOWN:
	    cerr << "wheel button down" << endl;
	    pCamera->Move( -pCamera->GetForwardVector()*5.0f);
	    break;
	  case SDL_BUTTON_LEFT:
	    vOld[GSE_Vector2::X] = event.button.x;
	    vOld[GSE_Vector2::Y] =  480.0f - event.button.y;
	    bMousePressed = 1;
	    cerr << "pressed" << endl;
	    break;
	  default:
	    break;
	  }
	}
	break;
	/// When mouse button is raised.
      case SDL_MOUSEBUTTONUP:
	{
	  switch( event.button.button )
	  {
	  case SDL_BUTTON_LEFT:
	    cerr << "released" << endl;
	    bMousePressed = 0;
	    break;
	  default:
	    break;
	  }
	}
	break;
      default:
	break;
      } // switch ( event
    } // while( POLL
    movementTimer.SetStartTime( SDL_GetTicks());
    bEventsFound = 0;
    nTimeLatter = SDL_GetTicks();
    nPassedTime = nTimeLatter - nTimeFormer;
    /////////////////////////////////////////////////////////////////
    /// Update animations every 5 milliseconds.
    if ( nPassedTime > 4 )
    {
      pSceneGraph->UpdateAnimations(nPassedTime);
      pSceneGraph->UpdateTransforms();    
      pSceneGraph->UpdateBoundingVolumes();
      nTimeFormer = SDL_GetTicks();
    }
    /// Collect render graph using current camera settings.
    pSceneGraph->CollectRenderGraph(pRenderer, pCameraNode->GetRenderCameraNode());
    /// Clear buffers, is such command is faced.
    pRenderer->SetClearBuffers( 1 );
    /////////////////////////////////////////////////////////////////
    /// Render only opaque polygons.
    pRenderer->SetRenderPass( OGL_RENDER_PASS_ONLY_OPAQUE );
    Graph::TravelDF<RenderNode::NodeType,GSE_OglRenderer>( pRoot, pRenderer );
    /////////////////////////////////////////////////////////////////
    /// Render only transparent polygons.
    /// DO NOT clear buffers, is such command is faced.
    pRenderer->SetClearBuffers( 0 );
    pRenderer->SetRenderPass( OGL_RENDER_PASS_ONLY_TRANSPARENT );
    Graph::TravelDF<RenderNode::NodeType,GSE_OglRenderer>( pRoot, pRenderer );
    /////////////////////////////////////////////////////////////////
    /// Swap buffers.
    pRenderer->SwapBuffers();
    /////////////////////////////////////////////////////////////////
    /// Clean up render graph.
    pSceneGraph->CleanUp();
    /////////////////////////////////////////////////////////////////
    /// one frame has been rendered.
    fpsCounter++;
    fpsCounter.SetCurrentTime(SDL_GetTicks());
    /////////////////////////////////////////////////////////////////
    /// Print FPS every second.
    if ( fpsCounter.HasPassedMS(1000))
    {
      cerr << "FPS: " << fpsCounter << endl;
      fpsCounter.ResetFrames();
      fpsCounter.SetStartTime( SDL_GetTicks());
      cerr << "collected objects: " << pSceneGraph->GetSceneRenderer().GetCollectedObjectCount() << endl;

    }
    /////////////////////////////////////////////////////////////////
    /// Select idle animation if necessary.
    pDwarf->CheckAnimationState();
  }
  return 0;
}
