/******************************************************************
 *   Copyright(c) 2006,2007 eNtity/Anssi Grhn
 * 
 *   This file is part of GSE.
 *
 *   GSE is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *    GSE is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with GSE; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 ******************************************************************/

#include "GSE_Frustum.h"
#include "GSE_geometry.h"
#include <list>

using std::list;
using Geometry::GSE_Plane;
using Geometry::GSE_PlaneIntersection;

GSE_Frustum::GSE_Frustum()
{

  // Add planes to the list 
  GSE_Plane top,bottom,left,right,near,far;

  AddPlane( top );
  AddPlane( bottom );
  AddPlane( left );
  AddPlane( right );
  AddPlane( near );
  AddPlane( far );

  // Make pointers to the planes
  list<GSE_Plane>::iterator it = Planes().begin();
  
  m_aPlanes[TOP]    = &(*it); it++;
  m_aPlanes[BOTTOM] = &(*it); it++;  
  m_aPlanes[LEFT]   = &(*it); it++;  
  m_aPlanes[RIGHT]  = &(*it); it++;  
  m_aPlanes[NEAR]   = &(*it); it++;  
  m_aPlanes[FAR]    = &(*it); 

}
void
GSE_Frustum::SetPlane(GSE_FrustumPlane nIndex, GSE_Plane &vPlane)
{
  *(m_aPlanes[nIndex]) = vPlane;
}
GSE_Plane &
GSE_Frustum::GetPlane( GSE_FrustumPlane nIndex )
{
  return *(m_aPlanes[nIndex]);
}
GSE_Vector3   
GSE_Frustum::GetCorner( GSE_FrustumCorner nIndex )
{
  return m_aCorners[nIndex];
}
void
GSE_Frustum::SetCorner( GSE_FrustumCorner nIndex, GSE_Vector3 vPoint )
{
  m_aCorners[nIndex] = vPoint;  
}
/////////////////////////////////////////////////////////////////
GSE_Frustum::GSE_FrustumIntersection_t
GSE_Frustum::IntersectsAABB( const GSE_AxisAlignedBox &aaBox )
{
  
  Geometry::GSE_PlaneIntersection iRel;

  for(int iPlane=0;iPlane<NUM_FRUSTUM_PLANES;iPlane++){

    iRel = Geometry::PlaneIntersectsAABB( GetPlane( (GSE_FrustumPlane)iPlane ), aaBox);
    // If the object is behind of any of the planes, 
    // then it is outside the frustum
    if (iRel == Geometry::BEHIND)
    {
      return OUTSIDE;
    }
    // If the object intersects any of the planes, 
    // then it is intersecting the frustum
    if (iRel == Geometry::INTERSECTS)
    {
      return INTERSECTION;
    }
  }  
  // if we end up here, the object is neither behind or intersecting of any of the 
  // planes. Hence, it is inside.
  return INSIDE;

}
/////////////////////////////////////////////////////////////////
GSE_Frustum::GSE_FrustumIntersection_t
GSE_Frustum::IntersectsSphere( const GSE_Sphere &sphere )
{
  
  Geometry::GSE_PlaneIntersection iRel;
  float fTmp = 0.0f;

  for(int iPlane=0;iPlane<NUM_FRUSTUM_PLANES;iPlane++){

    iRel = Geometry::SphereIntersectsPlane( GetPlane( (GSE_FrustumPlane)iPlane ), sphere, fTmp);
   
    // If the object is behind of any of the planes, 
    // then it is outside the frustum
    if (iRel == Geometry::BEHIND){
#ifdef DEBUG
      Core::GSE_Logger::Error() << "Sphere outside plane " << GetPlane((GSE_FrustumPlane)iPlane) ;
      switch((GSE_FrustumPlane)iPlane)
      {
      case TOP:
	Core::GSE_Logger::Error() << "TOP" << endl;
	break;
      case BOTTOM:
	Core::GSE_Logger::Error() << "BOTTOM" << endl;
	break;
      case LEFT:
	Core::GSE_Logger::Error() << "LEFT" << endl;
	break;
      case RIGHT:
	Core::GSE_Logger::Error() << "RIGHT" << endl;
	break;
      case NEAR:
	Core::GSE_Logger::Error() << "NEAR" << endl;
	break;
      case FAR:
	Core::GSE_Logger::Error() << "FAR" << endl;
	break;
      }
#endif
      return OUTSIDE;

    }
    // If the object intersects any of the planes, 
    // then it is intersecting the frustum
    if (iRel == Geometry::INTERSECTS) {
      return INTERSECTION;
    }
  }  
  // if we end up here, the object is neither behind or intersecting of any of the 
  // planes. Hence, it is inside.
  return INSIDE;
  
}


GSE_Frustum::GSE_FrustumIntersection_t
GSE_Frustum::IntersectsOrientedBox( const GSE_OrientedBox &obBox)
{
  
  char bUseLineSegmentTest = 0;
  
  // The effective radius of the box respect to a plane
  float fEffRadius	= 0.0f;
  float fDot		= 0.0f;
  float fDot2		= 0.0f;

  GSE_Vector3 vNormal;
  
  // When box length (principal axis) is much greater than the two others,
  // it is better to use line segment test for better accuracy.
  bUseLineSegmentTest = (( obBox.GetLength()-obBox.GetWidth() ) > (0.5f * obBox.GetWidth())) || 
                        ( (obBox.GetLength()-obBox.GetHeight()) > (0.5f * obBox.GetHeight()));
  
  if ( bUseLineSegmentTest ) 
  {

    // The line seqment
    GSE_Vector3 vQ1 = obBox.GetPosition() + (obBox.GetForwardVector()*(obBox.GetLength() * 0.5f)) ;
    GSE_Vector3 vQ2 = vQ1 - (obBox.GetForwardVector()*obBox.GetLength()) ;
    
    for(int iPlane=0;iPlane<NUM_FRUSTUM_PLANES;iPlane++){
      
      vNormal.UseExternalData( m_aPlanes[iPlane]->m_pValues);
      // Since axes from GSE_Orientable are always Unit length,
      // we include proper dimensions in the equation.
      fEffRadius = 0.5f * ( fabsf( (obBox.GetRightVector()*obBox.GetWidth()).Dot(vNormal) ) +
			    fabsf( (obBox.GetUpVector()*obBox.GetHeight()).Dot(vNormal) ) );
      
      // Calculate 4D dot product between a plane and the line endpoints
      fDot  = vNormal.Dot( vQ1) + (*m_aPlanes[iPlane])[GSE_Plane::D];
      fDot2 = vNormal.Dot( vQ2) + (*m_aPlanes[iPlane])[GSE_Plane::D];

      if (fDot <= -fEffRadius && fDot2 <= -fEffRadius)
      {
	return OUTSIDE;

      } else if ( (fDot < -fEffRadius && fDot2 > -fEffRadius)){
	// Cut off the part from the cylinder which lies outside the frustum
	float fT = (fEffRadius + fDot2) / vNormal.Dot(vQ2-vQ1) ;
	
	vQ1 = vQ2 + (fT*(vQ1-vQ2));

      } else if ( fDot2 < -fEffRadius && fDot > -fEffRadius ){
	
	// Cut off the part from the cylinder which lies outside the frustum
	float fT = (fEffRadius + fDot) / vNormal.Dot(vQ1-vQ2) ;
	vQ2 = vQ1 + (fT*(vQ2-vQ1));
      }
      
    }  
    
  } else {


    for(int iPlane=0;iPlane<NUM_FRUSTUM_PLANES;iPlane++){
      
      vNormal.UseExternalData( m_aPlanes[iPlane]->m_pValues);

      // Since axes from GSE_Orientable are always Unit length,
      // we include proper dimensions in the equation.
      
      fEffRadius = 0.5f * ( fabsf((obBox.GetForwardVector()*obBox.GetLength()).Dot(vNormal)) +
			   fabsf((obBox.GetRightVector()*obBox.GetWidth()).Dot(vNormal)) +
			   fabsf((obBox.GetUpVector()*obBox.GetHeight()).Dot(vNormal)) );
      
      // Calculate 4D dot product between plane and box center
      fDot  = vNormal.Dot( obBox.GetPosition()) + (*m_aPlanes[iPlane])[GSE_Plane::D];

      if ( fDot <= -fEffRadius )
      {
	return OUTSIDE;
      }
      
    }  
  }
  
  return INTERSECTION;
}

