/******************************************************************
 *   Copyright(c) 2006,2007 eNtity/Anssi Grhn
 * 
 *   This file is part of GSE.
 *
 *   GSE is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *    GSE is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with GSE; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 ******************************************************************/
/////////////////////////////////////////////////////////////////
#ifndef __GSE_ffmpeg_h__
#define __GSE_ffmpeg_h__
#include "GSE_Logger.h"
#include <list>
#include <ffmpeg/avcodec.h>
#include <ffmpeg/avformat.h>
#include <string>
/////////////////////////////////////////////////////////////////
using std::string;
using std::list;
using Core::GSE_Logger;
/////////////////////////////////////////////////////////////////
/// A Frame class which holds the ffmpeg AVFrame and the time it will be displayed on screen in seconds.
class FFMpegFrame 
{
 private:
  // The actual frame containing image information
  AVFrame *m_pFrame;		
  // Passed time that this frame has been shown
  unsigned int m_nFrameDisplayedMS; 
  unsigned int m_nFrameTime;
  unsigned int m_nCurrentFrame;
  

  
 public:
  ////////////////////
  /// The constructor
  FFMpegFrame()
  {
    m_pFrame = NULL;
    m_nFrameDisplayedMS = 0;
    m_nCurrentFrame = 0;
  }
  ////////////////////
  /// Destructor.
  ~FFMpegFrame()
  {
    if (m_pFrame != NULL ) av_free(m_pFrame);
    m_pFrame =NULL;
    m_nFrameDisplayedMS = 0;
  }
  ////////////////////
  /// Adds given time to display time.
  inline void AddToDisplayTimeMS( unsigned int nTimeMS )
  {
    m_nFrameDisplayedMS += nTimeMS;
  }
  ////////////////////
  /// Resets display time to 0.
  inline void ResetDisplayTime() 
  {
    m_nFrameDisplayedMS = 0;
  }
  ////////////////////
  /// How long this frame has been displayed, in milliseconds.
  inline unsigned int GetDisplayTimeMS()
  {
    return m_nFrameDisplayedMS;
  }
  ////////////////////
  /// Returns the AVFrame.
  inline AVFrame *GetAVFrame() const
  {
    return m_pFrame;
  }
  ////////////////////
  /// Sets the AVFrame.
  inline void SetAVFrame( AVFrame *pFrame ) 
  {
    m_pFrame = pFrame;
  }
  ////////////////////
  /// Returns the current frame number.
  /// \return frame number.
  inline unsigned int GetNumber() const
  {
    if ( m_pFrame != NULL ) return m_pFrame->pts;
    return 0;
  }

};
/////////////////////////////////////////////////////////////////
/// The class which decodes video stream using ffmpeg library.
class GSE_FFMpegStream 
{

 protected:
  /// Frame display time in milliseconds.
  unsigned int m_nFrameDisplayTimeMS;
  /// RGB image data.
  uint8_t *m_pBuffer;
  /// Stream filename.
  string m_strFilename;
  /// The format context for video.
  AVFormatContext *m_pFormatCtx;
  /// The codec context for video.
  AVCodecContext  *m_pCodecCtx;  
  /// The videostream which is decoded from the file.
  int m_iVideoStream;
  /// Should video be looped, default false.
  int m_bLooping;
  /// The frame which was returned by latest GetNextFrame() call.
  FFMpegFrame *m_pCurrentFrame;
  /// Retrieves the next frame from the video stream.
  int DecodeNextFrame( AVFrame *pFrame);
  /// Performs cleanup stuff.
  void Destroy();
  /// Initializes members variables.
  void Init();
  /// Subsection start.
  unsigned int m_nSectionStart;
  /// Subsection end.
  unsigned int m_nSectionEnd;
  /// Gets the time how long a frame must be displayed.
  /// \return frame display time in milliseconds.
  inline unsigned int GetFrameDisplayTimeMS() const 
  {
    return m_nFrameDisplayTimeMS;
  }
  /// Sets frame display time.
  /// \param nTimeMS frame display time in milliseconds.
  inline void SetFrameDisplayTimeMS( unsigned int nTimeMS )
  {
    m_nFrameDisplayTimeMS = nTimeMS;
  }
  ////////////////////
  /// Returns the next frame from the stream and deletes the previously returned one (if one exists). 
  /// When stream ends, NULL is returned.
  FFMpegFrame *	GetNextFrame();
  /// Is stream playing.
  int m_bIsPlaying;
  
 public:
  /// Default constructor.
  GSE_FFMpegStream();
  /// Destructor.
  ~GSE_FFMpegStream();
  /// Copy constructor.
  GSE_FFMpegStream( const GSE_FFMpegStream &ref);
  /// Assignment operator.
  /// \param obj ffmpeg stream to be assigned.
  GSE_FFMpegStream& operator=( GSE_FFMpegStream obj );
  ////////////////////
  /// Sets the video stream filename to be read.
  /// \param sFilename path to video file.
  int SetStream( char *sFilename );
  ////////////////////
  /// Returns the stream filename
  /// \return filename string.
  string GetStream() const;
  ////////////////////
  /// Has stream ended.
  // \return true if stream is in the end. (if looping, never true)
  int HasStreamEnded();
  ////////////////////
  /// Is stream looping.
  /// \return true if stream loops.
  int IsLooping();
  ////////////////////
  /// Sets the looping mode.
  /// \param bLooping true for enabling the loop, false for disabling.
  void SetLooping( int bLooping );
  ////////////////////
  /// Gets the pointer to the currently displayed frame.
  /// \return pointer to FFMpegFrame.
  FFMpegFrame *	GetCurrentFrame();
  ////////////////////
  /// Seeks frame nFrame in stream.
  /// \param nFrame frame position.
  void Seek( unsigned int nFrame );
  ////////////////////
  /// Fills buffer with image data.
  void UpdateBuffer();
  ////////////////////
  /// Returns image data buffer.
  /// \returns pointer to m_pBuffer.
  inline uint8_t *GetBuffer() { return m_pBuffer; }
  ////////////////////
  /// Seek next frame to be displayed.
  /// \param nPassedTimeMS forwards stream by that many milliseconds.
  void SeekNextFrameFromStream( unsigned int nPassedTimeMS );
  /// Sets subsection of the stream to be played.
  /// \param nTime Subsection start time milliseconds.
  /// \param nDuration Subsection duration in milliseconds.
  void SetSection( unsigned int nTime, unsigned int nDuration );
  /// Returns subsection start time.
  inline unsigned int GetSectionStart() const
  {
    return m_nSectionStart;
  }
  /// Returns subsection end time.
  inline unsigned int GetSectionEnd() const
  {
    return m_nSectionEnd;
  }
  /// Sets stream to playing mode.
  inline void Play() { m_bIsPlaying = 1; } 
  /// Stops playing.
  inline void Stop() { m_bIsPlaying = 0; } 
  /// Is stream playing now.
  /// \return true if so, false otherwise.
  inline int IsPlaying() const { return m_bIsPlaying; }
};
/////////////////////////////////////////////////////////////////
#endif
