/******************************************************************
 *   Copyright(c) 2006,2007 eNtity/Anssi Grhn
 * 
 *   This file is part of GSE.
 *
 *   GSE is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *    GSE is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with GSE; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 ******************************************************************/

/*
 * A class to make axis rotations easy
 *
 */
#include <math.h>
#include <iostream>
#include "GSE_Logger.h"
#include "GSE_math.h"
#include "GSE_debug.h"

using namespace Math;

GSE_Quaternion::GSE_Quaternion() : 
  m_vValues(0.0f,0.0f,0.0f), m_fW(0.0f)
{

}
GSE_Quaternion::GSE_Quaternion( float fX, float fY, float fZ, float fW) :
  m_vValues(fX,fY,fZ), m_fW(fW)
{

}
GSE_Quaternion::~GSE_Quaternion()
{
  //std::cerr << "Deleting Quaternion" << std::endl;
}

void 
GSE_Quaternion::CreateFromAxisAngle( float fX, float fY, float fZ, 
				     float fDegrees)
{
  
  CreateFromAxisAngleRad( fX, fY, fZ, (fDegrees / 180.0) * PI );
  
}

void 
GSE_Quaternion::CreateFromAxisAngle( GSE_Vector3 vValues, float fDegrees )
{
  CreateFromAxisAngleRad( vValues[GSE_Vector3::X], 
			  vValues[GSE_Vector3::Y], 
			  vValues[GSE_Vector3::Z], 
			  (fDegrees / 180.0f) * PI );
  
}

void 
GSE_Quaternion::CreateFromAxisAngleRad ( GSE_Vector3 vValues, float fRad )
{
  CreateFromAxisAngleRad( vValues[GSE_Vector3::X], 
			  vValues[GSE_Vector3::Y], 
			  vValues[GSE_Vector3::Z], fRad);
}

void 
GSE_Quaternion::CreateFromAxisAngleRad ( float fX, float fY, float fZ, 
					 float fRad ) 
{
  
  float thetaDiv2    = fRad * 0.5f;
  float sinThetaDiv2 = sinf ( thetaDiv2 );

  m_vValues[GSE_Vector3::X] = fX * sinThetaDiv2 ;
  m_vValues[GSE_Vector3::Y] = fY * sinThetaDiv2 ;
  m_vValues[GSE_Vector3::Z] = fZ * sinThetaDiv2 ;
  
  m_fW = cosf( thetaDiv2 );  

}
// -----------------------------------------------------------------
void 
GSE_Quaternion::CreateFromAxisAngles( GSE_Vector3 &vValues )
{
  GSE_Quaternion qX,qY,qZ,qCombined;

  qX.CreateFromAxisAngle( 1,0,0, vValues[GSE_Vector3::X]);
  qY.CreateFromAxisAngle( 0,1,0, vValues[GSE_Vector3::Y]);
  qZ.CreateFromAxisAngle( 0,0,1, vValues[GSE_Vector3::Z]);
  qCombined = qZ * qY * qX;

  *this = qCombined;
  
}
// -----------------------------------------------------------------
void
GSE_Quaternion::CreateFromAxisAnglesRad( GSE_Vector3 &vValues )
{
  GSE_Quaternion qX,qY,qZ,qCombined;

  qX.CreateFromAxisAngleRad( 1,0,0, vValues[GSE_Vector3::X]);
  qY.CreateFromAxisAngleRad( 0,1,0, vValues[GSE_Vector3::Y]);
  qZ.CreateFromAxisAngleRad( 0,0,1, vValues[GSE_Vector3::Z]);
  qCombined = qZ * qY * qX;
  
  *this = qCombined;
}
// -----------------------------------------------------------------
GSE_Matrix4x4f 
GSE_Quaternion::ToMatrix()
{
  
  GSE_Matrix4x4f mRotation;
  
  
// The values in Row-Major mode matrix which represents the 
// rotation matrix from the Quaternion:
// (From http://skal.planet-d.net/demo/matrixfaq.htm#Q2)
//
//                                                               
//           1 - (2Y  + 2Z ) 2XY - 2ZW           2XZ + 2YW     
//                                                               
//                                                               
//      M =  2XY + 2ZW         1 - (2X  + 2Z )   2YZ - 2XW     
//                                                               
//                                                               
//           2XZ - 2YW         2YZ + 2XW         1 - (2X  + 2Y)
//                                                               
//

  float xx = m_vValues.m_pValues[GSE_Vector3::X] * m_vValues.m_pValues[GSE_Vector3::X];
  float yy = m_vValues[GSE_Vector3::Y] * m_vValues[GSE_Vector3::Y];
  float zz = m_vValues[GSE_Vector3::Z] * m_vValues[GSE_Vector3::Z];
  
  float xy = m_vValues[GSE_Vector3::X] * m_vValues[GSE_Vector3::Y];
  float xz = m_vValues[GSE_Vector3::X] * m_vValues[GSE_Vector3::Z];
  float yz = m_vValues[GSE_Vector3::Y] * m_vValues[GSE_Vector3::Z];
  
  float wz = m_fW * m_vValues[GSE_Vector3::Z];
  float wy = m_fW * m_vValues[GSE_Vector3::Y];
  float wx = m_fW * m_vValues[GSE_Vector3::X];
  
  mRotation(0,0) = 1.0 - yy - yy - zz - zz;
  mRotation(0,1) = xy + xy - wz - wz;
  mRotation(0,2) = xz + xz + wy + wy;
  mRotation(0,3) = 0.0;
  
  mRotation(1,0) = xy + xy + wz + wz;
  mRotation(1,1) = 1.0f - xx - xx - zz - zz;
  mRotation(1,2) = yz + yz - wx - wx;
  mRotation(1,3) = 0.0;
  
  mRotation(2,0) = xz + xz - wy - wy;
  mRotation(2,1) = yz + yz + wx + wx;
  mRotation(2,2) = 1.0 - xx - xx - yy - yy;
  mRotation(2,3) = 0.0;
  
  mRotation(3,0) = 0.0;
  mRotation(3,1) = 0.0;
  mRotation(3,2) = 0.0;
  mRotation(3,3) = 1.0;
  
  return mRotation;

}

float &
GSE_Quaternion::operator[](Quaternion_Index iIndex )
{
  switch(iIndex){
  case GSE_Quaternion::W:
    return m_fW;
  default:
    return m_vValues[iIndex];
    
  }
}

GSE_Quaternion
GSE_Quaternion::operator*(GSE_Quaternion q) const
{

  GSE_Quaternion	qQuot;
  GSE_Vector3		vTmp;
  
  //
  // q1q2 = [w1w2 - v1.v2, w1v2 + w2v1 + v1^v2]
  //

  // The scalar component
  qQuot[GSE_Quaternion::W] = m_fW * q[GSE_Quaternion::W] - 
                             m_vValues.Dot(q.GetVector());
  // The Vector component
  vTmp     = (m_vValues * q[GSE_Quaternion::W]);
  vTmp     = vTmp + (q.GetVector() * m_fW);
  vTmp     = vTmp + m_vValues.Cross(q.GetVector());
  
#ifdef DEBUG
  std::cout.precision(12);
  std::cout << "vTmp 1): " << (m_vValues * q[GSE_Quaternion::W]) << std::endl;
  std::cout << "vTmp 2): " << (q.GetVector() * m_fW) << std::endl;
  std::cout << "vTmp 3): " <<  m_vValues.Cross(q.GetVector()) << std::endl;
#endif

  qQuot[GSE_Quaternion::X] = vTmp[GSE_Vector3::X];
  qQuot[GSE_Quaternion::Y] = vTmp[GSE_Vector3::Y];
  qQuot[GSE_Quaternion::Z] = vTmp[GSE_Vector3::Z];

  return qQuot;
  
}

GSE_Quaternion &
GSE_Quaternion::operator= (GSE_Quaternion qQuat)
{
  
  m_fW = qQuat[GSE_Quaternion::W];
  m_vValues[GSE_Vector3::X] = qQuat[GSE_Quaternion::X];
  m_vValues[GSE_Vector3::Y] = qQuat[GSE_Quaternion::Y];
  m_vValues[GSE_Vector3::Z] = qQuat[GSE_Quaternion::Z];
  
  return *this;
}

float
GSE_Quaternion::Magnitude() 
{
  return sqrt (  m_vValues[GSE_Vector3::X]* m_vValues[GSE_Vector3::X] + 
		 m_vValues[GSE_Vector3::Y]* m_vValues[GSE_Vector3::Y] + 
		 m_vValues[GSE_Vector3::Z]* m_vValues[GSE_Vector3::Z] + 
		 m_fW* m_fW );
}

GSE_Quaternion &
GSE_Quaternion::Normalize()
{
  float fMagnitudeInv = 1.0f/Magnitude();
  
  m_fW *= fMagnitudeInv;

  m_vValues[GSE_Vector3::X] *= fMagnitudeInv;
  m_vValues[GSE_Vector3::Y] *= fMagnitudeInv;
  m_vValues[GSE_Vector3::Z] *= fMagnitudeInv;

  return *this;
}


GSE_Quaternion 
GSE_Quaternion::Inverse() const
{

  GSE_Quaternion qInverted;
  
  qInverted[GSE_Quaternion::X] = m_vValues.m_pValues[GSE_Vector3::X];
  qInverted[GSE_Quaternion::Y] = m_vValues.m_pValues[GSE_Vector3::Y];
  qInverted[GSE_Quaternion::Z] = m_vValues.m_pValues[GSE_Vector3::Z];
  qInverted[GSE_Quaternion::W] = m_fW;

  // Convert to Unit Quaternion
  qInverted.Normalize();

  // Now inverse can be calculated by negating the vector part.
  qInverted[GSE_Quaternion::X] = -qInverted[GSE_Quaternion::X];
  qInverted[GSE_Quaternion::Y] = -qInverted[GSE_Quaternion::Y];
  qInverted[GSE_Quaternion::Z] = -qInverted[GSE_Quaternion::Z];

  return qInverted;
}

const GSE_Vector3 &
GSE_Quaternion::GetVector() const
{
  return m_vValues;
}

std::ostream & 
Math::operator<<(std::ostream &stream, Math::GSE_Quaternion qQuat)
{
  stream << "(" 
	 << qQuat[GSE_Quaternion::W] 
	 << ", (" 
	 << qQuat.GetVector() 
	 << "))";
  return stream;
}
// The comparison operator
int
GSE_Quaternion::operator==( GSE_Quaternion q ) 
{
  return (q[GSE_Quaternion::X] == m_vValues[GSE_Quaternion::X] &&
	  q[GSE_Quaternion::Y] == m_vValues[GSE_Quaternion::Y] &&
	  q[GSE_Quaternion::Z] == m_vValues[GSE_Quaternion::Z] &&
	  q[GSE_Quaternion::W] == m_fW);
}
// The dot product between this and qQuat
float
GSE_Quaternion::Dot( GSE_Quaternion qQuat )
{
  float fResult = 0.0;
  fResult += m_vValues[GSE_Quaternion::X] * qQuat[GSE_Quaternion::X];
  fResult += m_vValues[GSE_Quaternion::Y] * qQuat[GSE_Quaternion::Y];
  fResult += m_vValues[GSE_Quaternion::Z] * qQuat[GSE_Quaternion::Z];
  fResult += m_fW * qQuat.m_fW;

  return fResult;
}
GSE_Quaternion 
GSE_Quaternion::operator-() const
{
  return GSE_Quaternion(-m_vValues.m_pValues[GSE_Quaternion::X],
			-m_vValues.m_pValues[GSE_Quaternion::Y],
			-m_vValues.m_pValues[GSE_Quaternion::Z],
			-m_fW);
}

GSE_Quaternion 
GSE_Quaternion::operator+(GSE_Quaternion qQuat) const
{
  return GSE_Quaternion( m_vValues.m_pValues[GSE_Quaternion::X] + qQuat.m_vValues.m_pValues[GSE_Quaternion::X],
			 m_vValues.m_pValues[GSE_Quaternion::Y] + qQuat.m_vValues.m_pValues[GSE_Quaternion::Y],
			 m_vValues.m_pValues[GSE_Quaternion::Z] + qQuat.m_vValues.m_pValues[GSE_Quaternion::Z],
			 m_fW + qQuat.m_fW);
}

GSE_Quaternion 
GSE_Quaternion::operator-(GSE_Quaternion qQuat) const
{
  return GSE_Quaternion( m_vValues.m_pValues[GSE_Quaternion::X] - qQuat.m_vValues.m_pValues[GSE_Quaternion::X],
			 m_vValues.m_pValues[GSE_Quaternion::Y] - qQuat.m_vValues.m_pValues[GSE_Quaternion::Y],
			 m_vValues.m_pValues[GSE_Quaternion::Z] - qQuat.m_vValues.m_pValues[GSE_Quaternion::Z],
			 m_fW - qQuat.m_fW);
}

// The scaling operator
GSE_Quaternion 
GSE_Quaternion::operator*( float fValue ) const
{
  return GSE_Quaternion(m_vValues.m_pValues[GSE_Quaternion::X] * fValue,
			m_vValues.m_pValues[GSE_Quaternion::Y] * fValue,
			m_vValues.m_pValues[GSE_Quaternion::Z] * fValue,
			m_fW * fValue);
}


GSE_Quaternion
GSE_Quaternion::Slerp( GSE_Quaternion qFrom, GSE_Quaternion qTo, float fInterpolation )
{
  
  float fCos = qFrom.Dot(qTo);
  float fScaleFrom, fScaleTo;

  if( fCos < 0.0f ){
    fCos = -fCos;
    qTo = -qTo;
  }

  if ( 1.0f - fCos > 0.00001f ){
    float fOmega = acosf(fCos);
    float f1DivSinOmega = (1.0f/sinf(fOmega));
    fScaleFrom = sinf((1.0f - fInterpolation )*fOmega) * f1DivSinOmega;
    fScaleTo   = sinf( fInterpolation * fOmega )  * f1DivSinOmega;

  } else {
    fScaleFrom = 1.0f - fInterpolation;
    fScaleTo = fInterpolation;
  }
  GSE_Quaternion qSlerp;
  GSE_Quaternion tmp1 = (fScaleFrom * qFrom);
  GSE_Quaternion tmp2 = (fScaleTo * qTo);
  qSlerp =  tmp1 + tmp2;
  //Core::GSE_Logger::Error() << "tmp1" << tmp1 << std::endl;
  //Core::GSE_Logger::Error() << "tmp2" << tmp2 << std::endl;
  //Core::GSE_Logger::Error() << "tmp1+tmp2=" << (tmp1+tmp2) << std::endl;

  //Core::GSE_Logger::Error() << "qSlerp" <<qSlerp <<  std::endl;

  return qSlerp;
}
// The numerically stable (unless v0 == -v1 !) minimal rotation arc
// between vectors v0 and v1
GSE_Quaternion
GSE_Quaternion::RotationArc( GSE_Vector3 v0, GSE_Vector3 v1)
{

  GSE_Quaternion qRet;
  v0.ToUnit();
  v1.ToUnit();
  GSE_Vector3 vCross = v0.Cross(v1);
  float  fDot = v0.Dot(v1);
  float  fS   = sqrtf((1.0f+fDot)*2.0f);
  float  f1DivS = 1.0f / fS;

  qRet[GSE_Quaternion::X] = vCross[GSE_Vector3::X] * f1DivS;
  qRet[GSE_Quaternion::Y] = vCross[GSE_Vector3::Y] * f1DivS;
  qRet[GSE_Quaternion::Z] = vCross[GSE_Vector3::Z] * f1DivS;
  qRet[GSE_Quaternion::W] = fS * 0.5f;

  return qRet;
}

//
// -- GSE_Quaternion ends
//

// ---------------------------------------------------------
//
// The GSE_Matrix4x4f implementation
//
// ---------------------------------------------------------

GSE_Matrix<4,float> 
Math::RotationMatrix(const GSE_Vector3 & vAxis, float fRadians)
{

  return RotationMatrix( vAxis.m_pValues[0], vAxis.m_pValues[1], vAxis.m_pValues[2], fRadians );
  
}

GSE_Matrix<4,float> 
Math::RotationMatrix(float fX, float fY, float fZ, float fRadians)
{
  GSE_Matrix<4,float> mS;
  GSE_Matrix<4,float> uuT = GSE_Matrix<4,float>::Zero();
  GSE_Matrix<4,float> mResult;
  // An unit vector is required
  GSE_Vector3 vAxisUnit(fX,fY,fZ);
  vAxisUnit.ToUnit();
  // For readability 
  fZ = vAxisUnit[GSE_Vector3::Z];
  fX = vAxisUnit[GSE_Vector3::X];
  fY = vAxisUnit[GSE_Vector3::Y];
  // Sine and cosine of the fRadians
  float fSinAlpha = sinf(fRadians);
  float fCosAlpha = cosf(fRadians);

  // To remove couple of overlapping mulitiplications
  float fXX = fX * fX;
  float fXY = fX * fY;
  float fXZ = fX * fZ;
  float fYY = fY * fY;
  float fYZ = fY * fZ;
  float fZZ = fZ * fZ;
  // At first, we construct matrix S:
  // 
  //     ( 0 -z -y )
  // S = ( z  0 -x )
  //     (-y  x  0 )
  //
  mS(0,0) = 0.0; mS(0,1) = -fZ; mS(0,2) =  fY; mS(0,3) = 0.0;
  mS(1,0) =  fZ; mS(1,1) = 0.0; mS(1,2) = -fX; mS(1,3) = 0.0;
  mS(2,0) = -fY; mS(2,1) = -fX; mS(2,2) = 0.0; mS(2,3) = 0.0;
  mS(3,0) = 0.0; mS(3,1) = 0.0; mS(3,2) = 0.0; mS(3,3) = 1.0;
  // then sin (fRadians) * S
  mS = mS * fSinAlpha;
  // uuT = uT * u;
  uuT(0,0) = fXX; uuT(0,1)= fXY; uuT(0,2) = fXZ; 
  uuT(1,0) = fXY; uuT(1,1)= fYY; uuT(1,2) = fYZ; 
  uuT(2,0) = fXZ; uuT(2,1)= fYZ; uuT(2,2) = fZZ; 
  
  mResult = uuT + (GSE_Matrix<4,float>::Identity() - uuT)*fCosAlpha + mS;
  mResult(3,3) = 1.0;

  return mResult;
}
GSE_Matrix<4,float>
Math::RotationMatrix( const GSE_Vector3 &vRadians)
{
  return RotationMatrix( vRadians.m_pValues[GSE_Vector3::X], 
			 vRadians.m_pValues[GSE_Vector3::Y], 
			 vRadians.m_pValues[GSE_Vector3::Z]);
}
GSE_Matrix<4,float>
Math::RotationMatrix(float fX, float fY, float fZ )
{
  // Create quaternions from angles
  GSE_Quaternion qX,qY,qZ;

  qX.CreateFromAxisAngleRad( 1.0, 0.0, 0.0, fX);
  qY.CreateFromAxisAngleRad( 0.0, 1.0, 0.0, fY);
  qZ.CreateFromAxisAngleRad( 0.0, 0.0, 1.0, fZ);
  
  // combine quaternions and convert to matrix.
  return (qZ * qY * qX).ToMatrix();
}
/////////////////////////////////////////////////////////////////
GSE_Matrix<4,float> 
Math::UniformScaleMatrix( float fScale )
{
  GSE_Matrix<4,float> mScale;
  mScale(0,0) = mScale(1,1) = mScale(2,2) = fScale;
  mScale(3,3) = 1.0f;
  return mScale;
}
/////////////////////////////////////////////////////////////////
GSE_Matrix<4,float>
Math::TranslationMatrix(float fX, float fY, float fZ)
{
  
  GSE_Matrix<4,float> mResult = GSE_Matrix<4,float>::Identity();
  
  mResult(0,3) = fX;
  mResult(1,3) = fY;
  mResult(2,3) = fZ;
  
  return mResult;
}
GSE_Matrix<4,float> 
Math::TranslationMatrix(const GSE_Vector3 & vValues)
{
  return TranslationMatrix( vValues.m_pValues[GSE_Vector3::X], 
			    vValues.m_pValues[GSE_Vector3::Y],
 			    vValues.m_pValues[GSE_Vector3::Z] );
}
// ---------------------------------------------------------------------------------
//
// Calculates the tangent, bitangent and normal vectors in tangent space
// for given vertices vPoint0, vPoint1, vPoint2
//
// ---------------------------------------------------------------------------------
GSE_Matrix<3,float> 
Math::GetTBNMatrix( GSE_Vector3 vPoint0, GSE_Vector3 vPoint1, GSE_Vector3 vPoint2,
		    GSE_Vector2 vTexCoord0, GSE_Vector2 vTexCoord1, GSE_Vector2 vTexCoord2)
{
  GSE_Matrix<3,float> mResult;
    
  GSE_Vector3 vQ1 = vPoint1 - vPoint0;
  GSE_Vector3 vQ2 = vPoint2 - vPoint0;
  GSE_Vector2 vQTex1 = vTexCoord1 - vTexCoord0;
  GSE_Vector2 vQTex2 = vTexCoord2 - vTexCoord0;
  //  ----------------------------------------------------------
  // The formula for retrieving the tangent space matrix 
  //
  //   Q1.xyz = V1.xyz - V0.xyz
  //   Q2.xyz = V2.xyz - V0.xyz
  //   Q1.s   = V1.s - V0.s
  //   Q2.s   = V2.s - V0.s
  //   
  //
  //  | Tx Ty Tz |              1               |  Q2.t -Q1.t | | Q1.x Q1.y Q1.z |
  //  |          | =    ------------------    * |             |*|                |
  //  | Bx By Bz |     Q1.s*Q2.t - Q2.s*Q1.t    | -Q2.s  Q1.s | | Q2.x Q2.y Q2.z |
  //
  //  ----------------------------------------------------------
  float fDenominator = (vQTex1[GSE_Vector2::X] * vQTex2[GSE_Vector2::Y]) - 
                       (vQTex2[GSE_Vector2::X] * vQTex1[GSE_Vector2::Y]);
  float fCoefficient = 0.0;

  if ( TOO_CLOSE_TO_ZERO(fCoefficient))
  {
    // Handle error
    return GSE_Matrix<3,float>::Identity();

  } 
  else 
  {

    GSE_Vector3 vNormal(0.0f, 0.0f, 1.0f);
    fCoefficient = 1.0f / fDenominator;

    GSE_Vector3 vTangent( vQTex2[GSE_Vector2::Y] * vQ1[GSE_Vector3::X] +
			  -vQTex1[GSE_Vector2::Y] * vQ2[GSE_Vector3::X],
			  vQTex2[GSE_Vector2::Y] * vQ1[GSE_Vector3::Y] +
			  -vQTex1[GSE_Vector2::Y] * vQ2[GSE_Vector3::Y],
			  vQTex2[GSE_Vector2::Y] * vQ1[GSE_Vector3::Z] +
			  -vQTex1[GSE_Vector2::Y] * vQ2[GSE_Vector3::Z] );

    GSE_Vector3 vBitangent( -vQTex2[GSE_Vector2::X] * vQ1[GSE_Vector3::X] +
			    vQTex1[GSE_Vector2::X] * vQ2[GSE_Vector3::X],
			    -vQTex2[GSE_Vector2::X] * vQ1[GSE_Vector3::Y] +
			    vQTex1[GSE_Vector2::X] * vQ2[GSE_Vector3::Y],
			    -vQTex2[GSE_Vector2::X] * vQ1[GSE_Vector3::Z] +
			    vQTex1[GSE_Vector2::X] * vQ2[GSE_Vector3::Z]) ;

    vTangent   *= fCoefficient;
    vBitangent *= fCoefficient;
    // Ensure that the vectors are orthogonal
    vTangent = vTangent - ( vNormal.Dot(vTangent) * vNormal);    
    vBitangent = vBitangent - ( vNormal.Dot(vBitangent) * vNormal ) - 
		 ( vTangent.Dot(vBitangent)*vTangent );
    // construct the matrix
    mResult(0,0) = vTangent[GSE_Vector3::X];
    mResult(0,1) = vTangent[GSE_Vector3::Y];
    mResult(0,2) = vTangent[GSE_Vector3::Z];
    mResult(1,0) = vBitangent[GSE_Vector3::X];
    mResult(1,1) = vBitangent[GSE_Vector3::Y];
    mResult(1,2) = vBitangent[GSE_Vector3::Z];
    mResult(2,0) = vNormal[GSE_Vector3::X];
    mResult(2,1) = vNormal[GSE_Vector3::Y];
    mResult(2,2) = vNormal[GSE_Vector3::Z];
    
  }
  
  
  return mResult;
}

// Returns the determinant of the 4x4 float matrix
float 
Math::Det(GSE_Matrix<4,float> mMatrix)
{
  float fDet = 0.0;
  fDet += mMatrix(0,0) * Det(RemoveRowAndColumn<4,float>(mMatrix, 1,1));
  fDet -= mMatrix(0,1) * Det(RemoveRowAndColumn<4,float>(mMatrix, 1,2));
  fDet += mMatrix(0,2) * Det(RemoveRowAndColumn<4,float>(mMatrix, 1,3));
  fDet -= mMatrix(0,3) * Det(RemoveRowAndColumn<4,float>(mMatrix, 1,4));
  return fDet;
}
// Returns the determinant of the 3x3 float matrix
float 
Math::Det(GSE_Matrix<3,float> mMatrix)
{
  float fDet = 0.0;
  fDet += mMatrix(0,0) * Det(RemoveRowAndColumn<3,float>(mMatrix, 1,1));
  fDet -= mMatrix(0,1) * Det(RemoveRowAndColumn<3,float>(mMatrix, 1,2));
  fDet += mMatrix(0,2) * Det(RemoveRowAndColumn<3,float>(mMatrix, 1,3));
  return fDet;
}
// Returns the determinant of the 2x2 float matrix
float 
Math::Det(GSE_Matrix<2,float> mMatrix)
{
  float fDet = mMatrix(0,0) * mMatrix(1,1) -
	       mMatrix(0,1) * mMatrix(1,0);
  return fDet;
}

GSE_Matrix<3,float>
Math::CovarianceMatrix( VertexArray pVertexArray, unsigned int nNumVertices)
{
  
  GSE_Matrix<3,float> mCovariance;
  GSE_Vector3 vAveragePos(0.0f,0.0f,0.0f);
  // Calculate average position
  for(unsigned int iVertComponent = 0;iVertComponent < nNumVertices*3; iVertComponent+=3)
  {
    vAveragePos[GSE_Vector3::X] += pVertexArray[iVertComponent];
    vAveragePos[GSE_Vector3::Y] += pVertexArray[iVertComponent+1];
    vAveragePos[GSE_Vector3::Z] += pVertexArray[iVertComponent+2];
  }

  vAveragePos /= nNumVertices;

  // Calculate Covariance matrix
  for(unsigned int iVertComponent = 0;iVertComponent < nNumVertices*3; iVertComponent+=3)
  {
    float fTmpX = pVertexArray[iVertComponent]   - vAveragePos[GSE_Vector3::X];
    float fTmpY = pVertexArray[iVertComponent+1] - vAveragePos[GSE_Vector3::Y];
    float fTmpZ = pVertexArray[iVertComponent+2] - vAveragePos[GSE_Vector3::Z];

    mCovariance(0,0) += fTmpX * fTmpX;
    mCovariance(1,1) += fTmpY * fTmpY;
    mCovariance(2,2) += fTmpZ * fTmpZ;
    mCovariance(0,1) = mCovariance(1,0) += fTmpX * fTmpY;
    mCovariance(0,2) = mCovariance(2,0) += fTmpX * fTmpZ;
    mCovariance(1,2) = mCovariance(2,1) += fTmpY * fTmpZ;

  }    
  mCovariance /= nNumVertices;

  return mCovariance;
}
/////////////////////////////////////////////////////////////////
GSE_Matrix<3,float>
Math::CovarianceMatrix( VertexArray pVertexArray, const GSE_IndexBuffer &indexBuffer)
{
  
  GSE_Matrix<3,float> mCovariance;
  GSE_Vector3 vAveragePos(0.0f,0.0f,0.0f);
  unsigned int nVertexIndex;
  // Calculate average position
  for(unsigned int nIndex = 0;nIndex < indexBuffer.m_nNumIndices; nIndex++)
  {
    nVertexIndex = indexBuffer.m_pIndices[nIndex] * 3;
    vAveragePos[GSE_Vector3::X] += pVertexArray[nVertexIndex];
    vAveragePos[GSE_Vector3::Y] += pVertexArray[nVertexIndex+1];
    vAveragePos[GSE_Vector3::Z] += pVertexArray[nVertexIndex+2];
  }

  vAveragePos /= indexBuffer.m_nNumIndices;
  float fTmpX;
  float fTmpY;
  float fTmpZ;
  // Calculate Covariance matrix
  for(unsigned int nIndex = 0;nIndex < indexBuffer.m_nNumIndices; nIndex++)
  {
    nVertexIndex = indexBuffer.m_pIndices[nIndex] * 3;
    fTmpX = pVertexArray[nVertexIndex]   - vAveragePos[GSE_Vector3::X];
    fTmpY = pVertexArray[nVertexIndex+1] - vAveragePos[GSE_Vector3::Y];
    fTmpZ = pVertexArray[nVertexIndex+2] - vAveragePos[GSE_Vector3::Z];
    
    mCovariance(0,0) += fTmpX * fTmpX;
    mCovariance(1,1) += fTmpY * fTmpY;
    mCovariance(2,2) += fTmpZ * fTmpZ;
    mCovariance(0,1) = mCovariance(1,0) += fTmpX * fTmpY;
    mCovariance(0,2) = mCovariance(2,0) += fTmpX * fTmpZ;
    mCovariance(1,2) = mCovariance(2,1) += fTmpY * fTmpZ;

  }    
  mCovariance /= indexBuffer.m_nNumIndices;

  return mCovariance;
}
/////////////////////////////////////////////////////////////////
#define SIGN(VAL) (VAL < 0.0F ? -1.0F : 1.0F)
void 
Math::CalculateEigensystem( GSE_Matrix<3,float> &mMatrix, 
			    float &fLambda1, float &fLambda2, float &fLambda3,
			    GSE_Matrix<3,float> &mRes)
{
  const float fEpsilon = 1.0e-10F;
  const int nMaxSweeps = 32;

  float m11 = mMatrix(0,0);
  float m12 = mMatrix(0,1);
  float m13 = mMatrix(0,2);
  float m22 = mMatrix(1,1);
  float m23 = mMatrix(1,2);
  float m33 = mMatrix(2,2);

  mRes = GSE_Matrix<3,float>::Identity();

  for(int a=0;a<nMaxSweeps;a++)
  {
    //Core::GSE_Logger::Log() << "loop" << a << std::endl;
    // Exit if off-diagonal entries are small enough
    if ( (fabs(m12) < fEpsilon) && 
	 (fabs(m13) < fEpsilon) &&
	 (fabs(m23) < fEpsilon)) break;
    
    if ( m12 != 0.0F)
    {

      float u = (m22 - m11 ) * 0.5F / m12;
      float u2 = u * u;
      float u2p1 = u2 + 1.0F;
      float t = (u2p1 != u2 ) ? SIGN(u) * (sqrt(u2p1) - fabs(u)) : 0.5F / u;
					   
      float c = 1.0F / sqrt(t * t + 1.0F);
      float s = c * t;


      
      m11 -= t * m12;
      m22 += t * m12;
      m12 =  0.0 ;

      float temp = (c * m13) - (s * m23);
      m23 = (s * m13) + (c * m23);
      m13 = temp;      

      for(int i=0;i<3;i++)
      {
	float tmp = (c * mRes(i,0)) - (s * mRes(i,1));
	mRes(i,1) = (s * mRes(i,0)) + (c * mRes(i,1));
	mRes(i,0) = tmp;
      }
      
    }
    
    if ( m13 != 0.0F )
    {

      float u = (m33 - m11 ) * 0.5F / m13;
      float u2 = u * u;
      float u2p1 = u2 + 1.0F;
      float t = (u2p1 != u2 ) ? SIGN(u) * (sqrt(u2p1) - fabs(u)) : 0.5F / u;
      
      float c = 1.0F / sqrt((t * t) + 1.0F);
      float s = c * t;


      m11 -= t * m13;
      m33 += t * m13;
      m13  = 0.0;

      float temp = (c * m12) - (s * m23);
      m23 = (s * m12) + (c * m23);
      m12 = temp;

      for(int i=0;i<3;i++)
      {
	float tmp = (c * mRes(i,0)) - (s * mRes(i,2));
	mRes(i,2) = (s * mRes(i,0)) + (c * mRes(i,2));
	mRes(i,0) = tmp;
      }
    }

    if ( m23 != 0.0F )
    {
      float u = (m33 - m22 ) * 0.5F / m23;
      float u2 = u * u;
      float u2p1 = u2 + 1.0F;
      float t = ( u2p1 != u2 ) ? SIGN(u) * (sqrt(u2p1) - fabs(u)) : 0.5F / u;
      
      float c = 1.0F / sqrt(t * t + 1.0F);
      float s = c * t;
      

      m22 -= t * m23;
      m33 += t * m23;
      m23  = 0.0;
      
      float temp = (c * m12) - (s * m13);
      m13 = (s * m12) + (c * m13);
      m12 = temp;

      for(int i=0;i<3;i++)
      {
	float tmp = (c * mRes(i,1)) - (s * mRes(i,2));
	mRes(i,2) = (s * mRes(i,1)) + (c * mRes(i,2));
	mRes(i,1) = tmp;
      }
    }
  } //  for(int a=0;a<nMaxSweeps;a++)
  
  fLambda1 = m11;
  fLambda2 = m22;
  fLambda3 = m33;
 
}
//
// Copies values from the vVector to iRowth row of the matrix mMatrix
//
void
Math::SetRowVector(GSE_Matrix<3,float> &mMatrix, unsigned int iRow, GSE_Vector3 vVector)
{

  float *pArray = mMatrix.GetArray();
  unsigned int nIndex = (iRow % 3) * 3;
  
  pArray[nIndex]   = vVector[GSE_Vector3::X];
  pArray[nIndex+1] = vVector[GSE_Vector3::Y];
  pArray[nIndex+2] = vVector[GSE_Vector3::Z];

}
// Copies values from the vVector to iColth column of the matrix mMatrix
void 
Math::SetColumnVector(GSE_Matrix<3,float> &mMatrix, unsigned int iCol, GSE_Vector3 vVector)
{

  float *pArray = mMatrix.GetArray();
  unsigned int nIndex = iCol % 3;
  pArray[nIndex]   = vVector[GSE_Vector3::X];
  pArray[nIndex+3] = vVector[GSE_Vector3::Y];
  pArray[nIndex+6] = vVector[GSE_Vector3::Z];
  
}

GSE_Vector3 
Math::GetRowVector(GSE_Matrix<3,float> mMatrix, unsigned int iRow)
{
  return GSE_Vector3(mMatrix(iRow, 0),
		     mMatrix(iRow, 1),
		     mMatrix(iRow, 2));
}
GSE_Vector3 
Math::GetColumnVector(GSE_Matrix<3,float> mMatrix, unsigned int iCol)
{
  
  return GSE_Vector3(mMatrix(0, iCol),
		     mMatrix(1, iCol),
		     mMatrix(2, iCol));


}

GSE_Vector3 
Math::Translate(GSE_Vector3 v, GSE_Matrix4x4f m)
{
  return GSE_Vector3( v[GSE_Vector3::X]+m(0,3),
		      v[GSE_Vector3::Y]+m(1,3),
		      v[GSE_Vector3::Z]+m(2,3) );

}

GSE_Vector3 
Math::TranslateInverse(GSE_Vector3 v, GSE_Matrix4x4f m)
{
  return GSE_Vector3(v[GSE_Vector3::X]-m(0,3),
		     v[GSE_Vector3::Y]-m(1,3),
		     v[GSE_Vector3::Z]-m(2,3));
}

GSE_Vector3 
Math::Rotate( const GSE_Vector3 & v, const GSE_Matrix4x4f &m)
{
#define TMP_X v.m_pValues[GSE_Vector3::X]
#define TMP_Y v.m_pValues[GSE_Vector3::Y]
#define TMP_Z v.m_pValues[GSE_Vector3::Z]
 
  return GSE_Vector3(m.At(0,0) * TMP_X + m.At(0,1) * TMP_Y + m.At(0,2) * TMP_Z,
		     m.At(1,0) * TMP_X + m.At(1,1) * TMP_Y + m.At(1,2) * TMP_Z,
		     m.At(2,0) * TMP_X + m.At(2,1) * TMP_Y + m.At(2,2) * TMP_Z);
#undef TMP_X
#undef TMP_Y
#undef TMP_Z
}
/////////////////////////////////////////////////////////////////
GSE_Vector3 
Math::Transform( const GSE_Vector3 &v, const GSE_Matrix4x4f &m)
{
#define T_X    v.m_pValues[GSE_Vector3::X]
#define T_Y    v.m_pValues[GSE_Vector3::Y]
#define T_Z    v.m_pValues[GSE_Vector3::Z]
#define m00  m.At(0,0)
#define m01  m.At(0,1)
#define m02  m.At(0,2)
#define m03  m.At(0,3)
#define m10  m.At(1,0)
#define m11  m.At(1,1)
#define m12  m.At(1,2)
#define m13  m.At(1,3)
#define m20  m.At(2,0)
#define m21  m.At(2,1)
#define m22  m.At(2,2)
#define m23  m.At(2,3)

  return GSE_Vector3(m00*T_X + m01*T_Y + m02*T_Z + m03,
		     m10*T_X + m11*T_Y + m12*T_Z + m13,
		     m20*T_X + m21*T_Y + m22*T_Z + m23 );
#undef T_X
#undef T_Y
#undef T_Z
#undef m00
#undef m01
#undef m02
#undef m03
#undef m10
#undef m11
#undef m12
#undef m13
#undef m20
#undef m21
#undef m22
#undef m23
}
/////////////////////////////////////////////////////////////////
GSE_Vector3 
Math::MultiplyFromRight( GSE_Matrix3x3f mMatrix, GSE_Vector3 &vVector)
{

  return GSE_Vector3( mMatrix(0,0) * vVector[GSE_Vector3::X] + 
		      mMatrix(0,1) * vVector[GSE_Vector3::Y] + 
		      mMatrix(0,2) * vVector[GSE_Vector3::Z],
		      
		      mMatrix(1,0) * vVector[GSE_Vector3::X] + 
		      mMatrix(1,1) * vVector[GSE_Vector3::Y] + 
		      mMatrix(1,2) * vVector[GSE_Vector3::Z],
		      
		      mMatrix(2,0) * vVector[GSE_Vector3::X] + 
		      mMatrix(2,1) * vVector[GSE_Vector3::Y] + 
		      mMatrix(2,2) * vVector[GSE_Vector3::Z] );

}
/////////////////////////////////////////////////////////////////
GSE_Quaternion 
Math::RotationMatrixToQuaternion( const GSE_Matrix<4,float> &mMatrix )
{
    
  GSE_Quaternion qRetval;
  float fTrace = mMatrix.Trace();
  float fS = 0.0f;
  float f1DivS = 0.0f;
  const GSE_Matrix4x4f &m = mMatrix;

  if ( fTrace > 0.000001f )  // Greater than zero
  {
    
    fS = 2.0f * sqrt(fTrace);
    f1DivS = 1.0f / fS;
    qRetval[GSE_Quaternion::W] = 0.25f * fS;
   
    qRetval[GSE_Quaternion::X] = ( m.At(2,1) - m.At(1,2) ) * f1DivS;
    qRetval[GSE_Quaternion::Y] = ( m.At(0,2) - m.At(2,0) ) * f1DivS;
    qRetval[GSE_Quaternion::Z] = ( m.At(1,0) - m.At(0,1) ) * f1DivS;
    
  } else {
    
    // determine column where diagonal element has the greatest value
    unsigned int nColumnWithGreatestValue = 0;
    
    if ( m.At(0,0) > m.At(1,1)) {
      if ( m.At(0,0) > m.At(2,2) ) {
	nColumnWithGreatestValue = 0;
      } else {
	nColumnWithGreatestValue = 2;
      }
    } else {

      if ( m.At(1,1) > m.At(2,2) ) {
	nColumnWithGreatestValue = 1;
      } else {
	nColumnWithGreatestValue = 2;
      }
    }

    switch( nColumnWithGreatestValue ){
    case 0:
      fS  = sqrtf( 1.0f + m.At(0,0) - m.At(1,1) - m.At(2,2) ) * 2.0f;
      qRetval[GSE_Quaternion::X] = 0.25f * fS;
      qRetval[GSE_Quaternion::Y] = (m.At(1,0) + m.At(0,1) ) * f1DivS;
      qRetval[GSE_Quaternion::Z] = (m.At(0,2) + m.At(2,0) ) * f1DivS;
      qRetval[GSE_Quaternion::W] = (m.At(2,1) - m.At(1,2) ) * f1DivS;
    case 1:
      fS  = sqrtf( 1.0f + m.At(1,1) - m.At(0,0) - m.At(2,2) ) * 2.0f;
      qRetval[GSE_Quaternion::X] = (m.At(1,0) + m.At(0,1) ) * f1DivS;
      qRetval[GSE_Quaternion::Y] = 0.25f * fS;
      qRetval[GSE_Quaternion::Z] = (m.At(2,1) + m.At(1,2) ) * f1DivS;
      qRetval[GSE_Quaternion::W] = (m.At(0,2) - m.At(2,0) ) * f1DivS;
      break;
    case 2:
      fS  = sqrtf( 1.0f + m.At(2,2) - m.At(0,0) - m.At(1,1) ) * 2.0f;
      qRetval[GSE_Quaternion::X] = (m.At(0,2) + m.At(2,0) ) * f1DivS;
      qRetval[GSE_Quaternion::Y] = (m.At(2,1) + m.At(1,2) ) * f1DivS;
      qRetval[GSE_Quaternion::Z] = 0.25f * fS;
      qRetval[GSE_Quaternion::W] = (m.At(1,0) + m.At(0,1) ) * f1DivS;
    default:
      break;
    }

  }
  return qRetval;
}
/////////////////////////////////////////////////////////////////
GSE_Vector3 
Math::RotateAroundPoint( const GSE_Vector3 & vVector, 
			 const GSE_Vector3 & vPoint, 
			 const GSE_Quaternion & ref )
{
  return (vVector - vPoint).GetRotated(ref) + vPoint;
}
/////////////////////////////////////////////////////////////////
// As described in MatrixFAQ:
// 
//     cos_angle  = qr -> qw;
//     angle      = acos( cos_angle ) * 2 * RADIANS;
//     sin_angle  = sqrt( 1.0 - cos_angle * cos_angle );
//     if ( fabs( sin_angle ) < 0.0005 )
//       sin_angle = 1;
//     axis -> vx = qr -> qx / sin_angle;
//     axis -> vy = qr -> qy / sin_angle;
//     axis -> vz = qr -> qz / sin_angle;
/////////////////////////////////////////////////////////////////
void
Math::QuaternionToRotationAxisAndAngle( GSE_Quaternion &qQuat, 
					GSE_Vector3 &vAxis,
					float &fAngleInDegrees )
{
  float fCosAngle = qQuat[GSE_Quaternion::W];
  float fSinAngle = sqrt( 1.0f - fCosAngle * fCosAngle );
  
  // If denominator is too close to zero
  if ( fabs(fSinAngle < 0.0005f)) fSinAngle = 1;
  // The axis 
  vAxis = qQuat.GetVector();
  vAxis /= fSinAngle;
  // The angle
  fAngleInDegrees = RAD2DEG(acosf( fCosAngle )) * 2.0f;

}
/////////////////////////////////////////////////////////////////
