/******************************************************************
 *   Copyright(c) 2006,2007 eNtity/Anssi Grhn
 * 
 *   This file is part of GSE.
 *
 *   GSE is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *    GSE is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with GSE; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 ******************************************************************/
/******************************************************************
 * TODO: Write SRT-transform class. 3d game engine design, pp. 144-145.
 *
 *
 ******************************************************************/ 
#ifndef __GSE_math_h__
#define __GSE_math_h__

#include "GSE_IndexBuffer.h"
#include "GSE_structures.h"
#include <stdlib.h>

#include <iostream>
#include <math.h>
#include "GSE_debug.h" 

#define REALLY_SMALL_FLOAT 0.0000001f

#define TOO_CLOSE_TO_ZERO(FLOAT_VALUE) ( fabs(FLOAT_VALUE)< REALLY_SMALL_FLOAT )
#define QUITE_CLOSE_TO( VALUE, ANOTHER_VALUE ) ( (VALUE - ANOTHER_VALUE) < REALLY_SMALL_FLOAT )

namespace Math {
  // The handy PI constants
  const float PI     = 3.14159265358979323846264338327950288419716939937508f;
  const float PIdiv2 = 1.57079632679489661923132169163975144209858469968754f;

  // declaration is required for the ToMatrix() method.
  // ---------------------------------------------------------------------------- 
  //
  // The matrix template class 
  //
  // ----------------------------------------------------------------------------
  template <std::size_t SIZE, typename TYPE>
  class GSE_Matrix {
  protected:
    // the matrix values; stored in either row-major or column-major mode
    TYPE m_aValues[SIZE*SIZE];
  public:
    
    // ----------------------------------------------------------------------
    // A static method which returns diagonal matrix with ones.
    static GSE_Matrix Identity()
    {
      GSE_Matrix mDiagonal;
      TYPE *pValues = mDiagonal.GetArray();
      for ( unsigned int i=0;i<SIZE*SIZE;i++)
      {
	pValues[i] = i % (SIZE + 1) ? 0.0 : 1.0;	

      }
      return mDiagonal;
    }
    // A static method which returns zero matrix
    static GSE_Matrix Zero()
    {
      GSE_Matrix mZero;
      TYPE *pValues = mZero.GetArray();
      for ( unsigned int i=0;i<SIZE*SIZE;i++)
      {
	pValues[i] = 0.0;
      }
      return mZero;
    }

    // The default constructor
    GSE_Matrix()
    {
      for ( unsigned int i=0;i<SIZE*SIZE;i++)
      {
	m_aValues[i] = 0;
      }
    
    }
    
    // The parametrized constructor
    GSE_Matrix( const TYPE aValues[SIZE*SIZE] )
    {
       for ( unsigned int i=0;i<SIZE*SIZE;i++)
       {
	 m_aValues[i] = aValues[i];
       }
    
    }
    // Returns a singledimensional array of floats in Row-Major mode.
    // The values of the first row are listed first,
    // then the second row, etc.
    TYPE *GetArray() 
    {
      return m_aValues;
    }
    // The effective operation of this * mMatrix
    // This multiplies mMatrix from the left side.
    GSE_Matrix operator*(GSE_Matrix mMatrix) const
    {
      GSE_Matrix mNew;

      unsigned int iRow,iCol,iTmp;
      TYPE *pValues = mMatrix.GetArray();
      TYPE *pResult = mNew.GetArray();

    
      //
      // The Code for Row-Major mode calculation
      //
      for( iRow = 0; iRow<SIZE; iRow++)
      {
	for( iCol = 0; iCol<SIZE; iCol++)
	{
	  pResult[(iRow*SIZE)+iCol ] = 0.0;
	  for( iTmp = 0; iTmp<SIZE; iTmp++)
	  {
	    pResult[ (iRow*SIZE) + iCol] += m_aValues[ (iRow*SIZE) + iTmp] * pValues[ (iTmp*SIZE) + iCol];
	  }
	}
      }
      return mNew;
    }
    /// The vector multiplying method
    GSE_Vector<float,SIZE> operator*(const GSE_Vector<float,SIZE> &vVector) const
    {
      GSE_Vector<float,SIZE> vResult;
      for(unsigned int r=0;r<SIZE;r++)
      {
	vResult[r] = 0.0f;
	for(unsigned int c=0;c<SIZE;c++) // <-- C++, get it? har har.  =)
	{
	  vResult[r] += At(r,c) * vVector.m_pValues[c];
	}
      }
      return vResult;
    }
    // The assignment operator
    GSE_Matrix &operator=(GSE_Matrix mMatrix)
    {
       TYPE *pValues = mMatrix.GetArray();
       if ( pValues != NULL )
       {
	 for ( unsigned int i=0;i<SIZE*SIZE;i++)
	 {
	   m_aValues[i] = pValues[i];
	 }
       }
       return *this;
    }
    // Adds to matrices together
    GSE_Matrix operator+( GSE_Matrix mMatrix)
    {
      GSE_Matrix sum;
      for( unsigned int iIndex = 0;iIndex < SIZE*SIZE;iIndex++)
      {
	sum.GetArray()[iIndex] = mMatrix.GetArray()[iIndex]+m_aValues[iIndex];
      }
      return sum;
    }
    // Subtracts mMatrix from this.
    GSE_Matrix operator-( GSE_Matrix mMatrix)
    {
      GSE_Matrix sub;
      for( unsigned int iIndex = 0;iIndex < SIZE*SIZE;iIndex++)
      {
	sub.GetArray()[iIndex] = m_aValues[iIndex] - mMatrix.GetArray()[iIndex];
      }
      return sub;
    }
    // negates the current matrix
    GSE_Matrix operator-()
    {
      GSE_Matrix neg;
      for( unsigned int iIndex = 0;iIndex < SIZE*SIZE;iIndex++)
      {
	neg.GetArray()[iIndex] = -m_aValues[iIndex] ;
      }
      return neg;

    }
    // Multiplies matrix with scalar value fValue
    GSE_Matrix operator*( TYPE fValue )
    {
      GSE_Matrix mul;
      for( unsigned int iIndex = 0;iIndex < SIZE*SIZE;iIndex++)
      {
	mul.GetArray()[iIndex] = m_aValues[iIndex] * fValue;
      }
      return mul;
    }
    // Divides matrix with scalar value fValue
    GSE_Matrix operator/( TYPE Value )
    {
      GSE_Matrix div;
      // check for division by zero (or close to zero)
      if ( TOO_CLOSE_TO_ZERO(Value))
      {
	Core::GSE_Logger::Error() << __PRETTY_FUNCTION__ 
				  << "Division by zero possible, returning zero matrix" 
				  << std::endl;
	div = GSE_Matrix::Zero();    
      } 
      else 
      {
	for( unsigned int iIndex = 0;iIndex < SIZE;iIndex++)
	{
	  div.GetArray()[iIndex] = m_aValues[iIndex] / Value;
	}

      }
      
      return div;
    }
    // Returns reference to the float value at the (logically ) located at the
    // iRow:th row and iCol:th column of the matrix. Accepts values 0 - SIZE-1.
    const TYPE At(unsigned int iRow, unsigned int iCol) const
    {
      return m_aValues[( (iRow%SIZE) *SIZE )+(iCol%SIZE)];
    }
    // Returns reference to the float value at the (logically ) located at the
    // iRow:th row and iCol:th column of the matrix. Accepts values 0 - SIZE-1.
    TYPE &operator()( unsigned int iRow, unsigned int iCol )
    {
      return m_aValues[((iRow%SIZE)*SIZE)+(iCol%SIZE)];
    }
    // A hack for accessing matrix elements by using matrix[ROW][COLUMN] syntax.
    // This works only in Row-Major mode and return NULL if called in Column-Major mode.
    // Returns pointer to the beginning of ROW, which is a float array.
    // No boundary checks. No warnings. Could make a mess. Looks cool.
    TYPE *operator[]( unsigned int iIndex )
    {
      return &(m_aValues[(iIndex%SIZE)*SIZE]);
    }
    // Prints the matrix
    friend std::ostream& operator<<(std::ostream &stream, GSE_Matrix mat)
    {
      unsigned int iIndex = 0;
  
      for( iIndex = 0; iIndex<SIZE*SIZE; iIndex++)
      {
	  
	if ( iIndex != 0 && !(iIndex % SIZE)){
	  stream << std::endl;
	}
	stream << mat.m_aValues[iIndex] << " ";

      }
      
      stream << std::endl;
      return stream;
    }
    // Transposes the Matrix, returns reference to itself
    GSE_Matrix &Transpose()
    {
       unsigned int iRow = 0;
       unsigned int iCol = 0;
       float fTmp = 0.0;

       for( iRow=0;iRow<SIZE;iRow++)
       {
	 for( iCol=iRow+1;iCol<SIZE;iCol++)
	 {
      
	   fTmp = m_aValues[(iCol*SIZE)+iRow];
	   m_aValues[(iCol*SIZE)+iRow] = m_aValues[(iRow*SIZE)+iCol];
	   m_aValues[(iRow*SIZE)+iCol] = fTmp;
	   
	 }
       }
       return *this;
    }
    // Returns the transposed matrix as a copy
    GSE_Matrix Transposed() const
    {
      return GSE_Matrix(*this).Transpose();
    }
    // Returns the sum of the diagonal elements
    inline float Trace() const
    {
      float fSum = 0.0f;
      for( unsigned int nRow=0;nRow<SIZE;nRow++)
      {
	fSum += m_aValues[(nRow*SIZE)+nRow];
      }
      return fSum;
    }
    
    GSE_Matrix &operator/=(float fDivider)
    {
   
      unsigned int iRow = 0;
      unsigned int iCol = 0;


      for( iRow=0;iRow<SIZE;iRow++)
      {
	for( iCol=0;iCol<SIZE;iCol++)
	{
	  m_aValues[(iRow*SIZE)+iCol] /= fDivider;
	}
      }
      return *this;
    }
    // Swaps the two rows
    GSE_Matrix &SwapRows(unsigned int iRow1, unsigned int iRow2)
    {
      // Sanity check for row numbers
      if ( iRow1 < SIZE && iRow2 < SIZE ){


	TYPE *tmpRow = new TYPE[SIZE];
	// copy iRow1th row to tmpRow
	for(unsigned int iCol=0;iCol<SIZE;iCol++){
	  tmpRow[iCol] = (*this)(iRow1, iCol);
	}
	// copy iRow2th row to iRow1th row
	for(unsigned int iCol=0;iCol<SIZE;iCol++){
	  (*this)(iRow1,iCol) = (*this)(iRow2, iCol);
	}
	// copy tmpRow row to iRow2
	for(unsigned int iCol=0;iCol<SIZE;iCol++){
	  (*this)(iRow2, iCol) = tmpRow[iCol];
	}
	delete tmpRow;
	
      } else {
	Core::GSE_Logger::Error() << __FUNCTION__ 
			    << "Faulty row numbers!" << std::endl;
      }
      return *this;
    }
    GSE_Matrix &DivideRowBy(unsigned int iRow, TYPE fDivider )
    {
      for(unsigned int iCol=0;iCol<SIZE;iCol++){
	(*this)(iRow,iCol) /= fDivider;
      }
      return *this;
    }
    GSE_Matrix &MultiplyRowBy(unsigned int iRow, TYPE fMultiplier )
    {
      for(unsigned int iCol=0;iCol<SIZE;iCol++){
	(*this)(iRow,iCol) *= fMultiplier;
      }
      return *this;
    }
    GSE_Matrix &DivideColumnBy(unsigned int iCol, TYPE fDivider )
    {
      for(unsigned int iRow=0;iRow<SIZE;iRow++){
	(*this)(iRow,iCol) /= fDivider;
      }
      return *this;
    }
    GSE_Matrix &MultiplyColumnBy(unsigned int iCol, TYPE fMultiplier )
    {
      for(unsigned int iRow=0;iRow<SIZE;iRow++){
	(*this)(iRow,iCol) *= fMultiplier;
      }
      return *this;
    }
    // From http://www.devmaster.net/wiki/Matrix#Inverting_Matrices
    // 4x4 matrix inverse using Gauss-Jordan algorithm with row pivoting
    // originally written by Nathan Reed, now released into the public domain.
    GSE_Matrix Inverse()
    {

      GSE_Matrix<SIZE,TYPE> mOrig;    // The matrix whose inverse we want to take
      GSE_Matrix<SIZE,TYPE> mInverse;// initialize to 4x4 identity matrix
      mOrig = *this;
      mInverse = GSE_Matrix::Identity();
      
      //mat4 a(*this);         
      //mat4 b(identity);       
      unsigned int i, j, pivot;
      
      // loop through columns
      for (j = 0; j < SIZE; ++j){
	// select pivot element: maximum magnitude in this column
	for (pivot = 0, i = 1; i < SIZE; ++i){
	  if (fabs(mOrig(i,j)) > fabs(mOrig(pivot,j))){
	    pivot = i;
	  }
	}

	if ( TOO_CLOSE_TO_ZERO( mOrig(pivot,j)) )
	{
	  Core::GSE_Logger::Error() << "Matrix inverse does not exist!" 
			      << std::endl;
	  return GSE_Matrix<SIZE,TYPE>::Identity();
	}
	
	// interchange rows to put pivot element on the diagonal
	// skip if already on diagonal
	if (pivot != j){
	  
	  mOrig.SwapRows( pivot, j );
	  mInverse.SwapRows( pivot, j);

	  // ---
	  //vec4 temp = a.row(j);
	  //a.row(j) = a.row(pivot);
	  //a.row(pivot) = temp;
	  
	  //temp = b.row(j);
	  //b.row(j) = b.row(pivot);
	  //b.row(pivot) = temp;
	}
	
	// divide row by pivot element
	// skip if already equal to 1
	if ( mOrig(j,j) != 1.0f )
	{
	  
	  float temp = mOrig(j,j);
	  mOrig.DivideRowBy(j,temp);
	  mInverse.DivideRowBy(j,temp);
	  
	  //a.row(j) /= temp;
	  //b.row(j) /= temp;
	  
	  // Now the pivot element is 1
	}
	
	// subtract this row from others to make the rest of column j zero
	for (i = 0; i < SIZE; ++i)
	{
	  // skip rows that are already zero
	  if ((i != j) && !TOO_CLOSE_TO_ZERO( mOrig(i,j) ) )
	  {
	    float scale = -mOrig(i,j);
	    for(unsigned int iCol=0;iCol<SIZE;iCol++)
	    {
	      mOrig(i, iCol )    += mOrig(j,iCol)*scale;
	      mInverse(i, iCol ) += mInverse(j,iCol)*scale;
	    }
	    //a.row(i) += a.row(j) * scale;
	    //b.row(i) += b.row(j) * scale;
	  }
	}
      }
	
      // When these operations have been completed, a should have been transformed to the identity matrix
      // and b should have been transformed into the inverse of the original a
      return mInverse;
    }

  }; // template GSE_Matrix

  // class GSE_Matrix4x4f
  typedef GSE_Matrix<4,float> GSE_Matrix4x4f;
  // class GSE_Matrix3x3f
  typedef GSE_Matrix<3,float> GSE_Matrix3x3f;
  // class GSE_Matrix2x2f
  typedef GSE_Matrix<2,float> GSE_Matrix2x2f;
  // --------------------------------------------------------------------------
  //
  // Quaternion class 
  //
  // --------------------------------------------------------------------------
  class GSE_Quaternion {
  private:
    // The Quaternion components
    GSE_Vector3	m_vValues;
    float	m_fW; 
  public:
    typedef enum {
      X = 0,
      Y = 1,
      Z = 2,
      W = 3
    } Quaternion_Index;
    // The default constructor
    GSE_Quaternion();
    // The parametrized constructor.
    GSE_Quaternion( float fX, float fY, float fZ, float fW);
    // The deconstructor
    ~GSE_Quaternion();
    // Creates quaternion from axis angle in degrees
    void CreateFromAxisAngle ( float fX, float fY, float fZ, float fDegrees );
    // Creates quaternion from axis angle in degrees
    void CreateFromAxisAngle ( GSE_Vector3 vValues, float fDegrees );
    // Creates quaternion from axis angle in radians
    void CreateFromAxisAngleRad ( float fX, float fY, float fZ, float fRad );
    // Creates quaternion from axis angle in radians
    void CreateFromAxisAngleRad ( GSE_Vector3 vValues, float fRad );
    // Creates a quaternion from axis angles (degrees) with axis X=(1,0,0), Y=(0,1,0), Z=(0,0,1);
    void CreateFromAxisAngles( GSE_Vector3 &vValues );
    // Creates a quaternion from axis angles (radians) with axis X=(1,0,0), Y=(0,1,0), Z=(0,0,1);
    void CreateFromAxisAnglesRad( GSE_Vector3 &vValues );
    // Converts quaternion into rotation matrix
    GSE_Matrix4x4f ToMatrix();
    // The * Operator to enable multiplying Quaternions,
    // Effectively q1*q2 means first rotating by q2 and then by q1
    GSE_Quaternion operator * ( GSE_Quaternion qQuat ) const;
    // The assignment operator 
    GSE_Quaternion &operator = ( GSE_Quaternion qQuat );
    // The comparison operator
    int	operator==( GSE_Quaternion q );
    // The negation operator
    GSE_Quaternion operator-() const ;
    // The addition operator
    GSE_Quaternion operator+(GSE_Quaternion qQuat) const;
    // The subtraction operator
    GSE_Quaternion operator-(GSE_Quaternion qQuat) const;
    // The scaling operator
    GSE_Quaternion operator*( float value ) const;
    // So the scaling can be done both ways, from left or right.
    friend GSE_Quaternion operator*( float fValue, GSE_Quaternion qQuat )
    {
      GSE_Quaternion qScaled;
      qScaled[GSE_Quaternion::X] = qQuat[GSE_Quaternion::X] * fValue;
      qScaled[GSE_Quaternion::Y] = qQuat[GSE_Quaternion::Y] * fValue;
      qScaled[GSE_Quaternion::Z] = qQuat[GSE_Quaternion::Z] * fValue;
      qScaled[GSE_Quaternion::W] = qQuat[GSE_Quaternion::W] * fValue;
      return qScaled;
    }    
    // Returns the W,X,Y or Z component.
    float& operator[](Quaternion_Index iIndex );
    // Returns the magnitude of the Quaternion, square root of the 
    // sum of squares of each component.
    float Magnitude() ;
    // Normalizes the quaternion, ie. divides each component with the magnitude.
    GSE_Quaternion &Normalize();
    // Calculates the Quaternion Conjugate ie. The Inverse.
    GSE_Quaternion Inverse() const;
    // Reverses the quaternion (in rotation sense);
    inline GSE_Quaternion & Reverse()
    {
      m_fW = -m_fW;
      return *this;
    }
    inline GSE_Quaternion GetReverse() const
    {
      return GSE_Quaternion(*this).Reverse();
    }
    // Returns the Vector component 
    const GSE_Vector3 & GetVector() const;
    // For simplistic debug outputting
    friend std::ostream& operator<<(std::ostream &stream, GSE_Quaternion qQuat);
    // The dot product between this and qQuat
    float Dot( GSE_Quaternion qQuat );
    // -------------------------------------------------------------------------
    // Performs a spherical linear interpolation between quaternions qFrom and qTO
    static GSE_Quaternion Slerp(GSE_Quaternion qFrom, GSE_Quaternion qTo, float fInterpolation);
    // Calculates the rotation from v0 to v1
    static GSE_Quaternion RotationArc( GSE_Vector3 v0, GSE_Vector3 v1);
  };
  // -------------------------------------------------------------------------
  //
  // Other functions
  //
  // -------------------------------------------------------------------------
  // Generates a Rotation matrix which corresponds to rotation of fRadians rad
  // over axis (fX, fY, fZ)
  GSE_Matrix<4,float> RotationMatrix(float fX, float fY, float fZ, float fRadians);
  // Generates a Rotation matrix which corresponds to rotation of fRadians rad
  // over axis vAxis
  GSE_Matrix<4,float> RotationMatrix(const GSE_Vector3 &vAxis, float fRadians);
  // Generates a rotation matrix from (x,y,z) -axis rotations 
  // rotation fX radians using X-axis, fY rad over Y-axis and fZ rad over Z-axis.
  GSE_Matrix<4,float> RotationMatrix(float fX, float fY, float fZ );
  // Generates a rotation matrix from (x,y,z) -axis rotations 
  // rotation vRadians[X] radians using X-axis, 
  // vRadians[Y] rad over Y-axis and vRadians[Z] rad over Z-axis.
  GSE_Matrix<4,float> RotationMatrix(const GSE_Vector3 & vRadians );

  // Generates a translation matrix 
  GSE_Matrix<4,float> TranslationMatrix(float fX, float fY, float fZ);  
  GSE_Matrix<4,float> TranslationMatrix(const GSE_Vector3 &vValues);  
  
  GSE_Matrix<4,float> UniformScaleMatrix( float fScale );  
  // Constructs a Row-Major TBN (Tangent, Bitangent, Normal)-matrix for bump mapping
  GSE_Matrix<3,float> GetTBNMatrix( GSE_Vector3 vPoint0, GSE_Vector3 vPoint1, GSE_Vector3 vPoint2,
				    GSE_Vector2 vTexCoord0, GSE_Vector2 vTexCoord2, GSE_Vector2 vTexCoord3);

  // Returns the determinant of the 4x4 float matrix
  float	Det(GSE_Matrix<4,float> mMatrix);
  // Returns the determinant of the 3x3 float matrix
  float Det(GSE_Matrix<3,float> mMatrix);
  // Returns the determinant of the 2x2 float matrix
  float Det(GSE_Matrix<2,float> mMatrix);
  // Returns the iRowth row vector from 3x3 matrix.
  // iRow in range [0..SIZE-1 ]
  GSE_Vector3 GetRowVector(GSE_Matrix<3,float> mMatrix, unsigned int iRow);
  // Returns the iColth column vector from 3x3 matrix.
  // iCol in range [0..SIZE-1 ]
  GSE_Vector3 GetColumnVector(GSE_Matrix<3,float> mMatrix, unsigned int iCol);
  // Copies values from the vVector to iRowth row of the matrix mMatrix
  void SetRowVector(GSE_Matrix<3,float> &mMatrix, unsigned int iRow, GSE_Vector3 vVector);
  // Copies values from the vVector to iColth column of the matrix mMatrix
  void SetColumnVector(GSE_Matrix<3,float> &mMatrix, unsigned int iCol, GSE_Vector3 vVector);
  // Translates the vector v by matrix m ( doesn't use scaling or rotation )
  GSE_Vector3 Translate(GSE_Vector3 v, GSE_Matrix4x4f m);
  // Translates the vector v by the inverse of translation matrix m ( doesn't use scaling or rotation )
  GSE_Vector3 TranslateInverse(GSE_Vector3 v, GSE_Matrix4x4f m);
  // Rotates vector v by matrix m ( doesn't use scaling or translation )
  GSE_Vector3 Rotate( const GSE_Vector3 & v, const GSE_Matrix4x4f &m);
  // Transforms vector v by matrix (scaling, rotation and translation is applied)
  GSE_Vector3 Transform( const GSE_Vector3 &v, const GSE_Matrix4x4f &m);
  // Multiplies vector vVector from right by matrix mMatrix  ( mMatrix * vVector );
  GSE_Vector3 MultiplyFromRight( GSE_Matrix3x3f mMatrix, GSE_Vector3 &vVector);
  
  
  // Template function for getting submatrix from another matrix by removing 
  // iRowSkipth row and iColSkipth column from original.
  // Returns a square GSE_Matrix with dimensions one smaller than original.
  // iRowSkip and iColSkip values are in range [1..SIZE]
  template<std::size_t SIZE,typename TYPE> 
  GSE_Matrix<SIZE-1,TYPE> RemoveRowAndColumn(GSE_Matrix<SIZE,TYPE> mMatrix, 
					   unsigned int iRowSkip, 
					   unsigned int iColSkip)
  {

    GSE_Matrix<SIZE-1,TYPE> mResult;
    unsigned int iResultRow = 1;
    unsigned int iResultCol = 1;
    // Check that the skipped row is in this matrix
    if ( iRowSkip > SIZE ){
      Core::GSE_Logger::Error() << __PRETTY_FUNCTION__
				<< "Can't remove row which "
				<< "is not in the matrix! ("
				<< iRowSkip << " > " << SIZE << ")" << std::endl;
      return mResult;
    }
    // Check that the skipped column is in this matrix.
    if ( iColSkip > SIZE ){
      Core::GSE_Logger::Error() << __PRETTY_FUNCTION__
				<< "Can't remove column which "
				<< "is not in the matrix! ("
				<< iColSkip << " > " << SIZE << ")" << std::endl;
      return mResult;
    }
    // 
    for( unsigned int iRow = 1;iRow <= SIZE;iRow++)
    {
      if ( iRow == iRowSkip ){

      } else {
	iResultCol = 1;
	for( unsigned int iCol = 1;iCol <= SIZE;iCol++)
	{

	  if ( iCol == iColSkip ){
	    
	  } else {
	    /*Core::GSE_Logger::Log() << "inserting from " << iRow << "," << iCol << std::endl;
	      Core::GSE_Logger::Log() << "inserting at " << iResultRow << "," << iResultCol << std::endl;*/
	    mResult(iResultRow+1, iResultCol+1) = mMatrix(iRow+1,iCol+1);
	    iResultCol++;
	  }
	  
	}
	iResultRow++;
      }
    }
    return mResult;
  }
  // Returns the covariance matrix of the vertices in the pVertexArray
  GSE_Matrix<3,float> CovarianceMatrix( VertexArray pVertexArray, unsigned int nNumVertices );
  // Returns the covariance matrix of the vertices in the pVertexArray using the given indices 
  GSE_Matrix<3,float> CovarianceMatrix( VertexArray pVertexArray, 
					const GSE_IndexBuffer &indexBuffer );
  // Calculates the eigenvectors and eigenvalues for matrix mMatrix 
  // using the Jacobi method. For matrices with dimensions less than 10,
  // the algorithm is efficient.
  void CalculateEigensystem( GSE_Matrix<3,float> &mMatrix, 
			     float &fLambda1, float &fLambda2, float &fLambda3,
			     GSE_Matrix<3,float> &mEigenvectorMatrix);
  
  // Converts a rotation matrix into a quaternion.
  GSE_Quaternion RotationMatrixToQuaternion( const GSE_Matrix<4,float> &mMatrix );
  GSE_Vector3 RotateAroundPoint( const GSE_Vector3 & vVector, 
				 const GSE_Vector3 & vPoint, 
				 const GSE_Quaternion & ref );
  
#define DEG2RAD(A) (A*0.0174532925f)
#define RAD2DEG(R) (R*57.2957795201f)

  // Returns a random value between min and max (included)
  template <typename Type> 
  Type GetRandom( Type min, Type max )
  {
    Type val = max - min ;
    if ( min == max)
      return min;
    return  min + (Type)( (val *rand()) / (RAND_MAX+1) );
  }
  

  // Converts Quaternion qQuat into rotation axis vAxis and
  // and angle fAngle. vAxis is not normalized, ie. ||vAxis|| != 1.0f
  void QuaternionToRotationAxisAndAngle( GSE_Quaternion &qQuat, 
					 GSE_Vector3 &vAxis,
					 float &fAngle );
  
} // namespace Math

#endif
