#include <GSE.h>
#include <UnitTest++/UnitTest++.h>

TEST(CameraRotation)
{
  /////////////////////////////////////////////////////////////////
  // 1. Start: 
  //	       Camera pos = (0,0,0).
  //           fwd = (0,0,-1)
  //           up  = (0,1,0)
  //           right = (1,0,0);
  /////////////////////////////////////////////////////////////////
  GSE_Camera *pCamera = GSE_CameraMgr::GetInstance()->CreateCamera();
  pCamera->SetPosition( GSE_Vector3(0,0,0));
  // check 1.
  CHECK_ARRAY_CLOSE(GSE_Vector3(0,0,0).m_pValues,  pCamera->GetPosition().m_pValues, 3, 0.001f);
  CHECK_ARRAY_CLOSE(GSE_Vector3(0,0,-1).m_pValues, pCamera->GetForwardVector().m_pValues, 3, 0.001f);
  CHECK_ARRAY_CLOSE(GSE_Vector3(0,1,0).m_pValues, pCamera->GetUpVector().m_pValues, 3, 0.001f);
  CHECK_ARRAY_CLOSE(GSE_Vector3(1,0,0).m_pValues, pCamera->GetRightVector().m_pValues, 3, 0.001f);
  /////////////////////////////////////////////////////////////////
  // 2. Rotate  camera 90 degrees over positive y-axis. 
  //  
  // start:
  //	       Camera pos = (0,0,0).
  //           fwd        = (0,0,-1)
  //           up         = (0,1,0)
  //           right      = (1,0,0);
  // end:
  //	       Camera pos = (0,0,0).
  //           fwd        = (-1,0,0)
  //           up         = (0,1,0)
  //           right      = (0,0,-1);
  /////////////////////////////////////////////////////////////////
  GSE_Quaternion q;
  q.CreateFromAxisAngle(0,1,0,90.0f);
  GSE_Vector3 vPoint(0,0,0);
  pCamera->RotateAroundPoint( vPoint,q);
  // check 2.
  CHECK_ARRAY_CLOSE(GSE_Vector3(0,0,0).m_pValues,  pCamera->GetPosition().m_pValues, 3, 0.001f);
  CHECK_ARRAY_CLOSE(GSE_Vector3(-1,0,0).m_pValues, pCamera->GetForwardVector().m_pValues, 3, 0.001f);
  CHECK_ARRAY_CLOSE(GSE_Vector3( 0,1,0).m_pValues, pCamera->GetUpVector().m_pValues, 3, 0.001f);
  CHECK_ARRAY_CLOSE(GSE_Vector3( 0,0,-1).m_pValues, pCamera->GetRightVector().m_pValues, 3, 0.001f);

  /////////////////////////////////////////////////////////////////
  // 3. Lift camera 3 units along positive y-axis and rotate 90 degrees over positive y-axis,
  //    using that camera position as origo. 
  // start:
  //	       Camera pos = (0,0,0).
  //           fwd        = (-1,0,0)
  //           up         = (0,1,0)
  //           right      = (0,0,-1);
  // end:
  //	       Camera pos = (0,3,0).
  //           fwd        = (0,0,1)
  //           up         = (0,1,0)
  //           right      = (-1,0,0);
  pCamera->SetPosition(GSE_Vector3(0,3.0,0));
  vPoint = pCamera->GetPosition();
  pCamera->RotateAroundPoint( vPoint,q);
  // check 3.
  CHECK_ARRAY_CLOSE(GSE_Vector3(0,3,0).m_pValues,  pCamera->GetPosition().m_pValues, 3, 0.001f);
  CHECK_ARRAY_CLOSE(GSE_Vector3(0,0,1).m_pValues, pCamera->GetForwardVector().m_pValues, 3, 0.001f);
  CHECK_ARRAY_CLOSE(GSE_Vector3( 0,1,0).m_pValues, pCamera->GetUpVector().m_pValues, 3, 0.001f);
  CHECK_ARRAY_CLOSE(GSE_Vector3( -1,0,0).m_pValues, pCamera->GetRightVector().m_pValues, 3, 0.001f);

  /////////////////////////////////////////////////////////////////
  // 4. Rotate 90 degrees over positive x-axis, using camera position as origo.
  // start:
  //	       Camera pos = (0,3,0).
  //           fwd        = (0,0,1)
  //           up         = (0,1,0)
  //           right      = (-1,0,0);
  // end:
  //	       Camera pos = (0,3,0).
  //           fwd        = (0,-1,0)
  //           up         = (0,0,1)
  //           right      = (-1,0,0);
  q.CreateFromAxisAngle( 1.0,0.0,0.0, 90.0f);
  vPoint = pCamera->GetPosition();
  pCamera->RotateAroundPoint( vPoint, q);
  // Check 4
  CHECK_ARRAY_CLOSE(GSE_Vector3(0,3.0,0).m_pValues,  pCamera->GetPosition().m_pValues, 3, 0.001f);  
  CHECK_ARRAY_CLOSE(GSE_Vector3(0,-1,0).m_pValues,   pCamera->GetForwardVector().m_pValues, 3, 0.001f);
  CHECK_ARRAY_CLOSE(GSE_Vector3( 0,0,1).m_pValues, pCamera->GetUpVector().m_pValues, 3, 0.001f);
  CHECK_ARRAY_CLOSE(GSE_Vector3( -1,0,0).m_pValues, pCamera->GetRightVector().m_pValues, 3, 0.001f);

  /////////////////////////////////////////////////////////////////
  // 5. Set camera position as (12,0,0)
  // start:
  //	       Camera pos = (0,3,0).
  //           fwd        = (0,-1,0)
  //           up         = (0,0,1)
  //           right      = (-1,0,0);
  // end:
  //	       Camera pos = (12,0,0).
  //           fwd        = (0,-1,0)
  //           up         = (0,0,1)
  //           right      = (-1,0,0);
  pCamera->SetPosition(GSE_Vector3(12.0,0,0));
  
  CHECK_ARRAY_CLOSE(GSE_Vector3(12,0,0).m_pValues,  pCamera->GetPosition().m_pValues, 3, 0.001f);  
  CHECK_ARRAY_CLOSE(GSE_Vector3(0,-1,0).m_pValues,  pCamera->GetForwardVector().m_pValues, 3, 0.001f);
  CHECK_ARRAY_CLOSE(GSE_Vector3( 0,0,1).m_pValues, pCamera->GetUpVector().m_pValues, 3, 0.001f);
  CHECK_ARRAY_CLOSE(GSE_Vector3( -1,0,0).m_pValues, pCamera->GetRightVector().m_pValues, 3, 0.001f);

  /////////////////////////////////////////////////////////////////
  // 6. Set origo as (0,0,0), rotate 90 degrees over positive y axis. 
  // start:
  //	       Camera pos = (12,0,0).
  //           fwd        = (0,-1,0)
  //           up         = (0,0,1)
  //           right      = (-1,0,0);
  // end:
  //	       Camera pos = (0,0,-12).
  //           fwd        = (0,-1,0)
  //           up         = (1,0,0)
  //           right      = (0,0,1);
  vPoint[GSE_Vector3::X] =
  vPoint[GSE_Vector3::Y] =
  vPoint[GSE_Vector3::Z] = 0.0f;
  q.CreateFromAxisAngle( 0,1,0, 90.0f);  

  pCamera->RotateAroundPoint( vPoint, q);

  CHECK_ARRAY_CLOSE(GSE_Vector3(0,0,-12).m_pValues,  pCamera->GetPosition().m_pValues, 3, 0.001f);  
  CHECK_ARRAY_CLOSE(GSE_Vector3(0,-1,0).m_pValues,  pCamera->GetForwardVector().m_pValues, 3, 0.001f);
  CHECK_ARRAY_CLOSE(GSE_Vector3( 1,0,0).m_pValues, pCamera->GetUpVector().m_pValues, 3, 0.001f);
  CHECK_ARRAY_CLOSE(GSE_Vector3( 0,0,1).m_pValues, pCamera->GetRightVector().m_pValues, 3, 0.001f);
  
}

int main()
{
  return UnitTest::RunAllTests();

}
// ----------------------------------------------------------------- 
// $Log: GSE_cameraTest.cpp,v $
// Revision 1.3  2007/03/26 12:53:00  entity
// commented tests better
//
// Revision 1.2  2007/03/15 14:01:09  entity
// Removed references to Positional::Position()
//
// Revision 1.1  2007/03/10 18:35:32  entity
// initial import to cvs
//
// ----------------------------------------------------------------- 
